# Copyright (c) Acconeer AB, 2023
# All rights reserved

from __future__ import annotations

import copy
import enum
from typing import List, Optional

import attrs
import numpy as np
import numpy.typing as npt
from scipy.signal import filtfilt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    ENVELOPE_FWHM_M,
    AlgoParamEnum,
    AlgoProcessorConfigBase,
    ProcessorBase,
    find_peaks,
    get_distance_filter_coeffs,
    get_distance_filter_edge_margin,
    get_temperature_adjustment_factors,
    interpolate_peaks,
)


DEFAULT_SC_BG_NUM_STD_DEV = 6.0
DEFAULT_FIXED_THRESHOLD_VALUE = 100.0
DEFAULT_THRESHOLD_SENSITIVITY = 0.5


class MeasurementType(AlgoParamEnum):
    CLOSE_RANGE = enum.auto()
    FAR_RANGE = enum.auto()


class ProcessorMode(AlgoParamEnum):
    DISTANCE_ESTIMATION = enum.auto()
    LEAKAGE_CALIBRATION = enum.auto()
    RECORDED_THRESHOLD_CALIBRATION = enum.auto()


class ThresholdMethod(AlgoParamEnum):
    """Threshold methods.
    ``CFAR`` Constant False Alarm Rate.
    ``FIXED`` Fixed threshold.
    ``RECORDED`` Recorded threshold."""

    CFAR = enum.auto()
    FIXED = enum.auto()
    RECORDED = enum.auto()


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    processor_mode: ProcessorMode = attrs.field(
        default=ProcessorMode.DISTANCE_ESTIMATION, converter=ProcessorMode
    )
    threshold_method: ThresholdMethod = attrs.field(
        default=ThresholdMethod.CFAR, converter=ThresholdMethod
    )
    measurement_type: MeasurementType = attrs.field(
        default=MeasurementType.FAR_RANGE, converter=MeasurementType
    )
    threshold_sensitivity: float = attrs.field(default=DEFAULT_THRESHOLD_SENSITIVITY)
    fixed_threshold_value: float = attrs.field(default=DEFAULT_FIXED_THRESHOLD_VALUE)

    def _collect_validation_results(
        self, config: Optional[a121.SessionConfig]
    ) -> list[a121.ValidationResult]:
        return []


@attrs.frozen(kw_only=True)
class ProcessorContext:
    direct_leakage: Optional[npt.NDArray[np.complex_]] = attrs.field(default=None)
    phase_jitter_comp_ref: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    recorded_threshold_mean_sweep: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    recorded_threshold_noise_std: Optional[List[np.float_]] = attrs.field(default=None)
    bg_noise_std: Optional[List[float]] = attrs.field(default=None)
    reference_temperature: Optional[int] = attrs.field(default=None)
    loopback_peak_location_m: Optional[float] = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class ProcessorExtraResult:
    """
    Contains information for visualization in ET.
    """

    abs_sweep: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    used_threshold: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    distances_m: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)


@attrs.frozen(kw_only=True)
class ProcessorResult:
    estimated_distances: Optional[list[float]] = attrs.field(default=None)
    estimated_rcs: Optional[list[float]] = attrs.field(default=None)
    recorded_threshold_mean_sweep: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    recorded_threshold_noise_std: Optional[list[np.float_]] = attrs.field(default=None)
    direct_leakage: Optional[npt.NDArray[np.complex_]] = attrs.field(default=None)
    phase_jitter_comp_reference: Optional[npt.NDArray[np.float_]] = attrs.field(default=None)
    extra_result: ProcessorExtraResult = attrs.field(factory=ProcessorExtraResult)


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):
    """Distance processor

    For all used subsweeps, the ``profile`` and ``step_length`` must be the same.

    :param sensor_config: Sensor configuration
    :param metadata: Metadata yielded by the sensor config
    :param processor_config: Processor configuration
    :param subsweep_indexes:
        The subsweep indexes to be processed. If ``None``, all subsweeps will be used.
    :param context: Context
    """

    CFAR_GUARD_LENGTH_ADJUSTMENT = 4
    CFAR_WINDOW_LENGTH_ADJUSTMENT = 0.25

    # Standard deviation of angle in direct leakage over multiple sensor restarts.
    PHASE_JITTER_RESTART_STD = 0.05

    CLOSE_RANGE_LOOPBACK_IDX = 0
    CLOSE_RANGE_DIST_IDX = 1

    RLG_PER_HWAAS_MAP = {
        a121.Profile.PROFILE_1: 11.3,
        a121.Profile.PROFILE_2: 13.7,
        a121.Profile.PROFILE_3: 19.0,
        a121.Profile.PROFILE_4: 20.5,
        a121.Profile.PROFILE_5: 21.6,
    }

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_indexes: Optional[list[int]] = None,
        context: Optional[ProcessorContext] = None,
    ) -> None:
        if context is None:
            context = ProcessorContext()

        if subsweep_indexes is None:
            subsweep_indexes = list(range(sensor_config.num_subsweeps))

        if (
            processor_config.measurement_type is MeasurementType.CLOSE_RANGE
            and processor_config.processor_mode is not ProcessorMode.LEAKAGE_CALIBRATION
        ):
            if context.direct_leakage is None or context.phase_jitter_comp_ref is None:
                raise ValueError("Sufficient processor context not provided")

        # range_subsweep_indexes holds the subsweep indexes corresponding to range measurements.
        # - Far range - all subsweeps are range measurements.
        # - Close range - The first subsweep is used for phase jitter compensation and the while
        #   the second is a range measurement. The location of each is indicated by
        #   CLOSE_RANGE_LOOPBACK_IDX and CLOSE_RANGE_DIST_IDX.
        if processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            self._validate_close_config(sensor_config, subsweep_indexes)
            self.range_subsweep_indexes = [self.CLOSE_RANGE_DIST_IDX]
        elif processor_config.measurement_type == MeasurementType.FAR_RANGE:
            self.range_subsweep_indexes = subsweep_indexes
        self.range_subsweep_configs = self._get_subsweep_configs(
            sensor_config, self.range_subsweep_indexes
        )
        self._validate_range_configs(self.range_subsweep_configs)

        self.sensor_config = sensor_config
        self.metadata = metadata
        self.processor_config = processor_config
        self.context = context

        self.processor_config.validate(self.sensor_config)
        self.profile = self._get_profile(self.range_subsweep_configs)
        self.step_length = self._get_step_length(self.range_subsweep_configs)
        self.approx_step_length_m = self.step_length * APPROX_BASE_STEP_LENGTH_M
        self.start_point = self._get_start_point(self.range_subsweep_configs)
        self.num_points = self._get_num_points(self.range_subsweep_configs)

        self.base_step_length_m = self.metadata.base_step_length_m
        self.step_length_m = self.step_length * self.base_step_length_m

        self.filt_margin = get_distance_filter_edge_margin(self.profile, self.step_length)
        self.start_point_cropped = self.start_point + self.filt_margin * self.step_length
        self.num_points_cropped = self.num_points - 2 * self.filt_margin

        self.subsweep_bpts = self._get_subsweep_breakpoints(
            self.range_subsweep_configs, self.filt_margin
        )

        self.distances_m = (
            self.start_point_cropped + np.arange(self.num_points_cropped) * self.step_length
        ) * self.metadata.base_step_length_m

        (self.b, self.a) = get_distance_filter_coeffs(self.profile, self.step_length)

        self.processor_mode = processor_config.processor_mode
        self.threshold_method = processor_config.threshold_method
        self.num_stds_in_threshold = self._sensitivity_to_standard_deviations(
            self.processor_config.threshold_sensitivity
        )

        if self.processor_mode == ProcessorMode.DISTANCE_ESTIMATION:
            self._init_process_distance_estimation()
        elif self.processor_mode == ProcessorMode.LEAKAGE_CALIBRATION:
            pass
        elif self.processor_mode == ProcessorMode.RECORDED_THRESHOLD_CALIBRATION:
            self._init_recorded_threshold_calibration()
        else:
            raise RuntimeError

    @staticmethod
    def _sensitivity_to_standard_deviations(sensitivity: float) -> float:
        """Convert sensitivity to standard deviations used by recorded threshold and CFAR.

        0 sensitivity corresponds to 15 standard deviations. 1 sensitivity corresponds to 2
        standard deviations.
        """
        if sensitivity < 0.0 or 1.0 < sensitivity:
            raise ValueError("Sensitivity outside of valid interval(0.0 <= Sensitivity <= 1.0).")

        return 8.0 - 7.0 * sensitivity

    @classmethod
    def _get_subsweep_configs(
        cls, sensor_config: a121.SensorConfig, subsweep_indexes: list[int]
    ) -> list[a121.SubsweepConfig]:
        return [sensor_config.subsweeps[i] for i in subsweep_indexes]

    @classmethod
    def _get_profile(cls, subsweep_configs: list[a121.SubsweepConfig]) -> a121.Profile:
        profiles = {c.profile for c in subsweep_configs}

        if len(profiles) > 1:
            raise ValueError

        (profile,) = profiles
        return profile

    @classmethod
    def _get_step_length(cls, subsweep_configs: list[a121.SubsweepConfig]) -> int:
        step_lengths = {c.step_length for c in subsweep_configs}
        if len(step_lengths) > 1:
            raise ValueError
        (step_length,) = step_lengths
        return step_length

    @classmethod
    def _get_start_point(cls, subsweep_configs: list[a121.SubsweepConfig]) -> int:
        return subsweep_configs[0].start_point

    @classmethod
    def _get_num_points(cls, subsweep_configs: list[a121.SubsweepConfig]) -> int:
        return sum(c.num_points for c in subsweep_configs)

    @classmethod
    def _get_subsweep_breakpoints(
        cls, subsweep_configs: list[a121.SubsweepConfig], filt_margin: int
    ) -> list[int]:
        """Calculate the subsweep breakpoints of the full sweep(including filter margins).

        The breakpoints are calculated as the cumulative sum of the number of points in each
        subsweep. Then, the breakpoints are formed by adding filt_margin to the front
        of the list and subtracted from the last element in the list.
        """

        num_points_in_subsweeps = [
            subsweep_config.num_points for subsweep_config in subsweep_configs
        ]
        bpts = [
            sum(num_points_in_subsweeps[: idx + 1]) for idx in range(len(num_points_in_subsweeps))
        ]
        bpts.insert(0, filt_margin)
        bpts[-1] -= filt_margin
        return bpts

    @classmethod
    def _validate_range_configs(cls, subsweep_configs: list[a121.SubsweepConfig]) -> None:
        cls._validate_range(subsweep_configs)

        for c in subsweep_configs:
            if not c.phase_enhancement:
                raise ValueError

    @classmethod
    def _validate_range(cls, subsweep_configs: list[a121.SubsweepConfig]) -> None:
        step_length = cls._get_step_length(subsweep_configs)

        next_expected_start_point = None

        for c in subsweep_configs:
            if next_expected_start_point is not None:
                if c.start_point != next_expected_start_point:
                    raise ValueError

            next_expected_start_point = c.start_point + c.num_points * step_length

    @classmethod
    def _validate_close_config(
        cls, sensor_config: a121.SensorConfig, subsweep_indexes: List[int]
    ) -> None:
        ERROR_MSG = "Incorrect subsweep config for close range measurement"

        if subsweep_indexes != [cls.CLOSE_RANGE_LOOPBACK_IDX, cls.CLOSE_RANGE_DIST_IDX]:
            raise ValueError(ERROR_MSG)

        subsweep_configs = cls._get_subsweep_configs(sensor_config, subsweep_indexes)

        if (
            not subsweep_configs[cls.CLOSE_RANGE_LOOPBACK_IDX].enable_loopback
            and subsweep_configs[cls.CLOSE_RANGE_DIST_IDX].enable_loopback
        ):
            raise ValueError(ERROR_MSG)

    def process(self, result: a121.Result) -> ProcessorResult:
        range_subframes = [result.subframes[i] for i in self.range_subsweep_indexes]
        frame = np.concatenate(range_subframes, axis=1)
        if self.processor_config.measurement_type == MeasurementType.CLOSE_RANGE:
            lb_angle = np.angle(result.subframes[self.CLOSE_RANGE_LOOPBACK_IDX]).astype(float)
            if (
                self.processor_config.processor_mode
                == ProcessorMode.RECORDED_THRESHOLD_CALIBRATION
            ):
                lb_angle += self.PHASE_JITTER_RESTART_STD * self.num_stds_in_threshold

            if self.processor_mode != ProcessorMode.LEAKAGE_CALIBRATION:
                frame = self._apply_phase_jitter_compensation(self.context, frame, lb_angle)

        sweep = frame.mean(axis=0)
        filtered_sweep = filtfilt(self.b, self.a, sweep)
        abs_sweep = np.abs(filtered_sweep)
        abs_sweep = abs_sweep[self.filt_margin : -self.filt_margin]

        if self.processor_mode == ProcessorMode.DISTANCE_ESTIMATION:
            return self._process_distance_estimation(abs_sweep, result.temperature)
        elif self.processor_mode == ProcessorMode.LEAKAGE_CALIBRATION:
            return ProcessorResult(
                phase_jitter_comp_reference=lb_angle,
                direct_leakage=frame,
            )
        elif self.processor_mode == ProcessorMode.RECORDED_THRESHOLD_CALIBRATION:
            return self._process_recorded_threshold_calibration(abs_sweep)

        raise RuntimeError

    @staticmethod
    def _apply_phase_jitter_compensation(
        context: ProcessorContext,
        frame: npt.NDArray[np.complex_],
        lb_angle: npt.NDArray[np.float_],
    ) -> npt.NDArray[np.complex_]:
        if context.direct_leakage is None or context.phase_jitter_comp_ref is None:
            raise ValueError("Sufficient context not provided")

        adjusted_leakage = context.direct_leakage * np.exp(
            -1j * (context.phase_jitter_comp_ref - lb_angle)
        )

        return frame - adjusted_leakage  # type: ignore[no-any-return]

    def _init_process_distance_estimation(self) -> None:
        if self.threshold_method == ThresholdMethod.RECORDED:
            if (
                self.context.recorded_threshold_mean_sweep is None
                or self.context.recorded_threshold_noise_std is None
            ):
                raise ValueError("Missing recorded threshold inputs in context")
        elif self.threshold_method == ThresholdMethod.FIXED:
            self.threshold = np.full(
                self.num_points_cropped, self.processor_config.fixed_threshold_value
            )
        elif self.threshold_method == ThresholdMethod.CFAR:
            self.cfar_abs_noise = np.zeros(shape=self.num_points_cropped)
            if self.context.bg_noise_std is not None:
                for idx, tx_off_noise_std in enumerate(self.context.bg_noise_std):
                    self.cfar_abs_noise[
                        self.subsweep_bpts[idx] : self.subsweep_bpts[idx + 1]
                    ] = tx_off_noise_std
            self.cfar_margin = self.calc_cfar_margin(self.profile, self.step_length)
            self.window_length = self._calc_cfar_window_length(self.profile, self.step_length)
            self.guard_half_length = self._calc_cfar_guard_half_length(
                self.profile, self.step_length
            )

        if self.context.loopback_peak_location_m is not None:
            self.offset_m = calculate_offset(self.context.loopback_peak_location_m, self.profile)
        else:
            self.offset_m = 0.0

    @classmethod
    def _calc_cfar_window_length(cls, profile: a121.Profile, step_length: int) -> int:
        window_length_m = ENVELOPE_FWHM_M[profile] * cls.CFAR_WINDOW_LENGTH_ADJUSTMENT
        step_length_m = step_length * APPROX_BASE_STEP_LENGTH_M
        return max([1, int(window_length_m / step_length_m)])

    @classmethod
    def _calc_cfar_guard_half_length(cls, profile: a121.Profile, step_length: int) -> int:
        guard_length_m = ENVELOPE_FWHM_M[profile] * cls.CFAR_GUARD_LENGTH_ADJUSTMENT
        step_length_m = step_length * APPROX_BASE_STEP_LENGTH_M
        guard_half_length_m = guard_length_m / 2
        return int(guard_half_length_m / step_length_m)

    @classmethod
    def calc_cfar_margin(cls, profile: a121.Profile, step_length: int) -> int:
        return cls._calc_cfar_window_length(
            profile, step_length
        ) + cls._calc_cfar_guard_half_length(profile, step_length)

    def _process_distance_estimation(
        self, abs_sweep: npt.NDArray[np.float_], temperature: int
    ) -> ProcessorResult:
        self.threshold = self._update_threshold(abs_sweep, temperature)

        found_peaks_idx = find_peaks(abs_sweep, self.threshold)
        (estimated_distances, estimated_amplitudes) = interpolate_peaks(
            abs_sweep,
            found_peaks_idx,
            self.start_point_cropped,
            self.step_length,
            self.base_step_length_m,
        )

        if self.processor_config.threshold_method == ThresholdMethod.CFAR:
            cfar_margin_slice = slice(self.cfar_margin, -self.cfar_margin)
            extra_result = ProcessorExtraResult(
                abs_sweep=abs_sweep[cfar_margin_slice],
                used_threshold=self.threshold[cfar_margin_slice],
                distances_m=self.distances_m[cfar_margin_slice],
            )
        else:
            extra_result = ProcessorExtraResult(
                abs_sweep=abs_sweep, used_threshold=self.threshold, distances_m=self.distances_m
            )

        # Calculate rcs before applying offset as the offset could push the estimated distance
        # into the next subsweep, resulting in rcs being calculated with wring sensor parameters.
        esimated_rcs = self._convert_amplitudes_to_rcs(
            estimated_amplitudes,
            estimated_distances,
            self.range_subsweep_configs,
            self.context.bg_noise_std,  # type: ignore[arg-type]
        )

        estimated_distances = [dist - self.offset_m for dist in estimated_distances]
        return ProcessorResult(
            estimated_distances=estimated_distances,
            estimated_rcs=esimated_rcs,
            extra_result=extra_result,
        )

    def _init_recorded_threshold_calibration(self) -> None:
        self.bg_sc_mean = np.zeros(self.num_points_cropped)
        self.bg_sc_sum_squared_bg_sweeps = np.zeros(self.num_points_cropped)
        self.sc_bg_num_sweeps = 1.0

    def _process_recorded_threshold_calibration(
        self, abs_sweep: npt.NDArray[np.float_]
    ) -> ProcessorResult:
        min_num_sweeps_in_valid_threshold = 2

        self.bg_sc_mean += abs_sweep
        self.bg_sc_sum_squared_bg_sweeps += np.square(abs_sweep)
        mean_sweep = self.bg_sc_mean / self.sc_bg_num_sweeps
        mean_square = self.bg_sc_sum_squared_bg_sweeps / self.sc_bg_num_sweeps
        square_mean = np.square(mean_sweep)

        if min_num_sweeps_in_valid_threshold <= self.sc_bg_num_sweeps:
            sc_bg_sweep_std = np.sqrt(
                np.abs(mean_square - square_mean)
                * self.sc_bg_num_sweeps
                / (self.sc_bg_num_sweeps - 1)
            )

            assert self.context.bg_noise_std is not None
            recorded_threshold_noise_std = []
            for idx, tx_off_noise_std in enumerate(self.context.bg_noise_std):
                # Subtract filt_margin from breakpoint to transform from full to cropped sweep
                subsweep_slice = slice(
                    self.subsweep_bpts[idx] - self.filt_margin,
                    self.subsweep_bpts[idx + 1] - self.filt_margin,
                    1,
                )
                # Clamp negative values to zero as the variable is in square root below.
                sigma_square_diff = np.clip(
                    sc_bg_sweep_std[subsweep_slice] ** 2 - tx_off_noise_std**2,
                    a_min=0,
                    a_max=None,
                )
                recorded_threshold_noise_std.append(
                    np.mean(np.sqrt(sigma_square_diff) / mean_sweep[subsweep_slice])
                )
        else:
            recorded_threshold_noise_std = None

        self.sc_bg_num_sweeps += 1

        extra_result = ProcessorExtraResult(abs_sweep=abs_sweep)
        return ProcessorResult(
            extra_result=extra_result,
            recorded_threshold_mean_sweep=mean_sweep,
            recorded_threshold_noise_std=recorded_threshold_noise_std,
        )

    def update_config(self, config: ProcessorConfig) -> None:
        ...

    def _update_threshold(
        self, abs_sweep: npt.NDArray[np.float_], temperature: int
    ) -> npt.NDArray[np.float_]:
        if self.threshold_method == ThresholdMethod.CFAR:
            return self._calculate_cfar_threshold(
                abs_sweep,
                self.window_length,
                self.guard_half_length,
                self.num_stds_in_threshold,
                self.cfar_abs_noise,
            )
        elif self.threshold_method == ThresholdMethod.FIXED:
            return self.threshold
        elif self.threshold_method == ThresholdMethod.RECORDED:
            assert self.context.reference_temperature is not None
            assert self.context.bg_noise_std is not None
            (
                signal_adjustment_factor,
                noise_adjustment_factor,
            ) = get_temperature_adjustment_factors(
                temperature_diff=temperature - self.context.reference_temperature,
                profile=self.profile,
            )
            return self._update_recorded_threshold(
                self.context,
                self.subsweep_bpts,
                self.num_stds_in_threshold,
                self.filt_margin,
                signal_adjustment_factor,
                noise_adjustment_factor,
            )
        else:
            raise RuntimeError

    @classmethod
    def _update_recorded_threshold(
        cls,
        context: ProcessorContext,
        bpts: list[int],
        num_stds: float,
        filt_margin: int,
        signal_adjustment_factor: float,
        noise_adjustment_factor: float,
    ) -> npt.NDArray[np.float_]:
        """Updates the recorded threshold to account for temperature effects.

        The threshold is constructed by adding a number of standard deviations of the background
        noise(tx off) and the signal noise(calculated during calibration) to the mean sweep. The
        mean sweep is adjusted by a factor and the background noise is adjusted before fed to this
        method to account for the temperature effect.
        """

        assert context.recorded_threshold_mean_sweep is not None
        assert context.recorded_threshold_noise_std is not None
        assert context.bg_noise_std is not None

        threshold = copy.deepcopy(context.recorded_threshold_mean_sweep) / signal_adjustment_factor

        for idx, (std_tx_off, std_recorded_threshold) in enumerate(
            zip(context.bg_noise_std, context.recorded_threshold_noise_std)
        ):
            # Subtract filt_margin from breakpoint to transform from full to cropped sweep
            subsweep_slice = slice(bpts[idx] - filt_margin, bpts[idx + 1] - filt_margin)
            threshold[subsweep_slice] += (
                np.sqrt(
                    (
                        noise_adjustment_factor * std_tx_off**2
                        + threshold[subsweep_slice] ** 2 * std_recorded_threshold**2
                    )
                )
                * num_stds
            )

        return threshold

    @staticmethod
    def _calculate_cfar_threshold(
        abs_sweep: npt.NDArray[np.float_],
        window_length: int,
        guard_half_length: int,
        num_stds: float,
        abs_noise_std: npt.NDArray,
    ) -> npt.NDArray[np.float_]:
        """Calculate CFAR threshold.

        Each point of the threshold is formed by using data from neighbouring segments of the
        sweep.

        The distance between a point and the start of the neighbouring segment is determined by
        the guard half length. The width of the segment by the window length.

        For each point, the threshold is calculated as the average of the points located in the
        segments to the left and to the right.
        """

        threshold = np.full(abs_sweep.shape, np.nan)
        margin = window_length + guard_half_length
        sweep_len_without_margins = abs_sweep.shape[0] - 2 * margin

        filt_abs_sweep = np.convolve(abs_sweep, np.ones(window_length), "valid") / window_length
        threshold[margin:-margin] = (
            filt_abs_sweep[:sweep_len_without_margins]
            + filt_abs_sweep[-sweep_len_without_margins:]
        ) / 2

        threshold += abs_noise_std * num_stds
        return threshold

    @classmethod
    def _convert_amplitudes_to_rcs(
        cls,
        amplitudes: list[float],
        distances: list[float],
        subsweeps: list[a121.SubsweepConfig],
        bg_noise_std: list[float],
    ) -> list[float]:

        # Determine subsweep breakpoints in meters.
        start_points = [subsweep.start_point for subsweep in subsweeps]
        bpts_m = np.array(start_points) * APPROX_BASE_STEP_LENGTH_M

        rcs = []
        # Loop over amplitude/distance pairs.
        for amplitude, distance in zip(amplitudes, distances):

            subsweep_idx = np.sum(bpts_m < distance) - 1

            # For the current distance, get corresponding sensor parameters and background noise.
            sigma = bg_noise_std[subsweep_idx]
            hwaas = subsweeps[subsweep_idx].hwaas
            profile = subsweeps[subsweep_idx].profile
            step_length = subsweeps[subsweep_idx].step_length

            # Calculate rcs.
            processing_gain_db = 10 * np.log10(cls.calc_processing_gain(profile, step_length))
            s_db = 20 * np.log10(amplitude)
            n_db = 20 * np.log10(sigma)
            r_db = 40 * np.log10(distance)
            rlg_db = cls.RLG_PER_HWAAS_MAP[profile] + 10 * np.log10(hwaas)

            rcs.append(s_db - n_db - rlg_db + r_db - processing_gain_db)

        return rcs

    @staticmethod
    def calc_processing_gain(profile: a121.Profile, step_length: int) -> float:
        """
        Approximates the processing gain of the matched filter.
        """
        envelope_base_length_m = ENVELOPE_FWHM_M[profile] * 2  # approx envelope width
        num_points_in_envelope = (
            int(envelope_base_length_m / (step_length * APPROX_BASE_STEP_LENGTH_M)) + 2
        )
        mid_point = num_points_in_envelope // 2
        pulse = np.concatenate(
            (
                np.linspace(0, 1, mid_point),
                np.linspace(1, 0, num_points_in_envelope - mid_point),
            )
        )
        return float(np.sum(pulse**2))


def calculate_bg_noise_std(
    subframe: npt.NDArray[np.complex_], subsweep_config: a121.SubsweepConfig
) -> float:
    profile = subsweep_config.profile
    step_length = subsweep_config.step_length
    (B, A) = get_distance_filter_coeffs(profile, step_length)
    filt_margin = get_distance_filter_edge_margin(profile, step_length)

    sweep = subframe.squeeze(axis=0)
    filtered_sweep = filtfilt(B, A, sweep)
    abs_sweep = np.abs(filtered_sweep)
    abs_sweep = abs_sweep[filt_margin:-filt_margin]

    return float(np.sqrt(np.mean(np.square(abs_sweep))))


def calculate_offset(peak_location: float, profile: a121.Profile) -> float:

    # Intercept and offset term of offset compensation.
    OFFSET_COMPENSATION_COEFFS = {
        a121.Profile.PROFILE_1: (0.76520069, 0.01146756),
        a121.Profile.PROFILE_2: (0.88898859, 0.02202941),
        a121.Profile.PROFILE_3: (0.92195011, 0.01356246),
        a121.Profile.PROFILE_4: (0.90674059, 0.01593211),
        a121.Profile.PROFILE_5: (1.08015653, 0.00390931),
    }

    p = OFFSET_COMPENSATION_COEFFS[profile]
    return p[0] * peak_location + p[1]


def calculate_loopback_peak_location(result: a121.Result, config: a121.SensorConfig) -> float:
    (B, A) = get_distance_filter_coeffs(config.profile, config.step_length)
    sweep = np.squeeze(result.frame, axis=0)
    abs_sweep = np.abs(filtfilt(B, A, sweep))
    peak_idx = [int(np.argmax(abs_sweep))]

    (estimated_dist, _) = interpolate_peaks(
        abs_sweep=abs_sweep,
        peak_idxs=peak_idx,
        start_point=config.start_point,
        step_length=config.step_length,
        step_length_m=APPROX_BASE_STEP_LENGTH_M,
    )

    return estimated_dist[0]



#!/usr/bin/env python
"""
VIRTUAL COM PORT TERMINAL
- implements a read/write terminal for communicating
with USB CDC device on winusb driver

SERIAL STATE notifications (2 bytes, interrupt endpoint)
    15..7 - reserved
    6   bOverRun    Received data has been discarded due to a device overrun
    5   bParity     A parity error has occurred
    4   bFraming    A framing error has occurred
    3   bRingSignal State of the ring indicator (RI)
    2   bBreak      Break state
    1   bTxCarrier  State of the data set ready (DSR)
    0   bRxCarrier  State of carrier detect (CD)

Line Coding Data Field (7 bytes, control endpoint)
    offset field       (bytes) Description
    --------------------------------------------------------------------
    0      dwDTERate   4       bit rate (bits per second)
    4      bCharFormat 1       stop bits (0 : 1bit, 1, 1.5bits, 2, 2bits)
    5      bParityType 1       0:None, 1:Odd, 2:Even, 3:Mark, 4:Space
    6      bDataBits   1       5, 6, 7, 8, 16

Control Line State Field (2 bytes, control endpoint)
    wValueBit  Description     (2 bytes data)
    ---------------------------
    bit 1 = 0  RTS : de-assert (negative voltage)
    bit 1 = 1  RTS : assert    (positive voltage)
    bit 0 = 0  DTR : de-assert (negative voltage)
    bit 0 = 1  DTR : assert    (positive voltage)

"""

from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import os
import time

from .winusbclasses import UsbSetupPacket
from .winusbpy import WinUsbPy


CDC_CMDS = {
    "SEND_BREAK": 0x23,  # wValue is break time
}


def config_log():
    log = logging.getLogger("usb_cdc")
    if "PYTERMINAL_DEBUG" in os.environ:
        log.setLevel(logging.DEBUG)
        fileHandler = logging.FileHandler("terminal.log")
        log_fmt = logging.Formatter(
            "%(levelname)s %(name)s %(threadName)-10s " + "%(funcName)s() %(message)s"
        )
        fileHandler.setFormatter(log_fmt)
        log.addHandler(fileHandler)
    return log


log = config_log()


class ComPort:
    def __init__(self, serial=None, vid=None, pid=None, start=True):
        self.serial = serial
        self.vid = vid
        self.pid = pid

        if not (vid and pid):
            raise AttributeError("Must provide vid & pid of device to connect to")

        self.winusbpy = WinUsbPy()
        self._rxremaining = b""
        self.maximum_packet_size = 0
        self.timeout = 0.01
        self.is_open = False
        if start:
            self.open()

    def open(self):
        # Control interface
        device_list = self.winusbpy.find_all_devices()
        device_path = None
        for device in device_list:
            if self.vid == device["vid"] and self.pid == device["pid"]:
                if self.serial is not None:
                    if self.serial == device["serial"]:
                        device_path = device["path"]
                        break
                else:
                    device_path = device["path"]
                    break
        if device_path is None:
            return False

        if not self.winusbpy.init_winusb_device(device_path):
            return False

        # Data Interface
        self.winusbpy.change_interface(0)
        interface2_descriptor = self.winusbpy.query_interface_settings(0)

        pipe_info_list = map(
            self.winusbpy.query_pipe, range(interface2_descriptor.b_num_endpoints)
        )
        for item in pipe_info_list:
            if item.pipe_id & 0x80:
                self._ep_in = item.pipe_id
            else:
                self._ep_out = item.pipe_id
            self.maximum_packet_size = (
                min(item.maximum_packet_size, self.maximum_packet_size) or item.maximum_packet_size
            )

        self.is_open = True

        self.winusbpy.set_timeout(self._ep_in, self.timeout)
        self.reset_input_buffer()
        return True

    def read(self, size=None):
        if not self.is_open:
            return None
        rx = [self._rxremaining]
        length = len(self._rxremaining)
        self._rxremaining = b""
        end_timeout = time.monotonic() + (self.timeout or 0.2)
        if size:
            while length < size:
                c = self.winusbpy.read(self._ep_in, size - length)
                if c is not None and len(c):
                    rx.append(c)
                    length += len(c)
                if time.monotonic() > end_timeout:
                    break
        else:
            while True:
                c = self.winusbpy.read(self._ep_in, self.maximum_packet_size)
                if c is not None and len(c):
                    rx.append(c)
                    length += len(c)
                else:
                    break
                if time.monotonic() > end_timeout:
                    break
        chunk = b"".join(rx)
        if size and len(chunk) >= size:
            if self._rxremaining:
                self._rxremaining = chunk[size:] + self._rxremaining
            else:
                self._rxremaining = chunk[size:]
            chunk = chunk[0:size]
        return chunk

    def write(self, data):
        if not self.is_open:
            return None
        try:
            ret = self.winusbpy.write(self._ep_out, bytes(data))
        except Exception as e:
            log.warning("USB Error on write {}".format(e))
            return

        if len(data) != ret:
            log.error("Bytes written mismatch {0} vs {1}".format(len(data), ret))
        else:
            log.debug("{} bytes written to ep".format(ret))

    def send_break(self, duration=0.25):
        if not self.is_open:
            return None

        duration_s = int(duration * 1000)

        txdir = 0  # 0:OUT, 1:IN
        req_type = 1  # 0:std, 1:class, 2:vendor
        # 0:device, 1:interface, 2:endpoint, 3:other
        recipient = 1
        req_type = (txdir << 7) + (req_type << 5) + recipient

        pkt = UsbSetupPacket(
            request_type=req_type,
            request=CDC_CMDS["SEND_BREAK"],
            value=duration_s,
            index=0x00,
            length=0x00,
        )

        buff = None

        wlen = self.winusbpy.control_transfer(pkt, buff)
        log.debug("Send break, {}b sent".format(wlen))

    def disconnect(self):
        if not self.is_open:
            return None
        self.winusbpy.close_winusb_device()
        self.is_open = False

    def __del__(self):
        self.disconnect()

    def reset_input_buffer(self):
        if self.is_open:
            self.winusbpy.flush(self._ep_in)
            while self.read():
                pass
        self._rxremaining = b""

    def flush(self):
        if not self.is_open:
            return None
        self.winusbpy.flush(self._ep_in)

    def close(self):
        self.disconnect()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import argparse
import sys

import nox


py_ver = ".".join(map(str, sys.version_info[:2]))
nox.options.sessions = ["lint", "docs", f"test(python='{py_ver}')"]
nox.options.stop_on_first_error = True
nox.options.reuse_existing_virtualenvs = True


BLACK_SPEC = "black==22.12.0"
ISORT_SPEC = "isort==5.6.3"
MYPY_SPEC = "mypy==0.991"
PIP_SPEC = "pip>=21.3"
PYTEST_MOCK_SPEC = "pytest-mock==3.3.1"
PYTEST_SPEC = "pytest==7.2"
PYTEST_XDIST_SPEC = "pytest-xdist==3.1.0"

SPHINX_SOURCE_DIR = "docs"
SPHINX_OUTPUT_DIR = "docs/_build"
SPHINX_HTML_ARGS = ("-b", "html", SPHINX_SOURCE_DIR, SPHINX_OUTPUT_DIR)


class Parser(argparse.ArgumentParser):
    KNOWN_TEST_GROUPS = ["unit", "integration", "app"]
    DEFAULT_TEST_GROUPS = ["unit", "integration"]

    KNOWN_DOCS_BUILDERS = ["html", "latexpdf", "rediraffecheckdiff", "rediraffewritediff"]
    DEFAULT_DOCS_BUILDERS = ["html"]

    def __init__(self):
        super().__init__()

        self.add_argument(
            "-e",
            "--editable",
            action="store_true",
            default=False,
        )

        test_group = self.add_argument_group("test")
        test_group.add_argument(
            "--test-groups",
            nargs="+",
            choices=self.KNOWN_TEST_GROUPS,
            default=self.DEFAULT_TEST_GROUPS,
        )
        test_group.add_argument(
            "--integration-args",
            nargs=argparse.REMAINDER,
            default=["a111", "--mock"],
        )
        test_group.add_argument(
            "--pytest-args",
            nargs=argparse.REMAINDER,
        )

        docs_group = self.add_argument_group("docs")
        docs_group.add_argument(
            "--docs-builders",
            nargs="+",
            choices=self.KNOWN_DOCS_BUILDERS,
            default=self.DEFAULT_DOCS_BUILDERS,
        )


@nox.session
def lint(session):
    session.install(
        BLACK_SPEC,
        ISORT_SPEC,
        "flake8>=4,<5",
        "flake8-future-annotations",
        "flake8-mutable",
        "flake8-quotes",
        "flake8-tidy-imports",
        "packaging",
    )
    session.run("python", "tools/check_permissions.py")
    session.run("python", "tools/check_whitespace.py")
    session.run("python", "tools/check_line_length.py")
    session.run("python", "tools/check_sdk_mentions.py")
    session.run("python", "tools/check_changelog.py")
    session.run("python", "tools/check_copyright.py")
    session.run("python", "-m", "flake8")
    session.run("python", "-m", "black", "--check", "--diff", "--quiet", ".")
    session.run("python", "-m", "isort", "--check", "--diff", "--quiet", ".")


@nox.session
def reformat(session):
    session.install(
        BLACK_SPEC,
        ISORT_SPEC,
    )
    session.run("python", "-m", "black", ".")
    session.run("python", "-m", "isort", ".")


@nox.session
def mypy(session):
    args = Parser().parse_args(session.posargs)

    if args.editable:
        session.install("-e", ".")
    else:
        session.install(".")

    session.install(MYPY_SPEC, PYTEST_SPEC)
    session.run("python", "-m", "mypy", "-p", "acconeer.exptool.a121")
    session.run("python", "-m", "mypy", "-p", "acconeer.exptool.app.new")
    session.run("python", "-m", "mypy", "-p", "acconeer.exptool.setup")
    session.run("python", "-m", "mypy", "tests/unit/a121")


@nox.session
def docs(session):
    args = Parser().parse_args(session.posargs)

    if args.editable:
        session.install(PIP_SPEC)
        session.install("-e", ".[docs]")
    else:
        session.install(".[docs]")

    if "html" in args.docs_builders:
        session.run("python", "-m", "sphinx", "-W", *SPHINX_HTML_ARGS)

    if "latexpdf" in args.docs_builders:
        session.run(
            "python",
            "-m",
            "sphinx",
            "-M",
            "latexpdf",
            SPHINX_SOURCE_DIR,
            SPHINX_OUTPUT_DIR,
        )

    if "rediraffewritediff" in args.docs_builders:
        session.run(
            "python",
            "-m",
            "sphinx",
            "-b",
            "rediraffewritediff",
            SPHINX_SOURCE_DIR,
            SPHINX_OUTPUT_DIR,
        )

    if "rediraffecheckdiff" in args.docs_builders:
        session.run(
            "python",
            "-m",
            "sphinx",
            "-b",
            "rediraffecheckdiff",
            SPHINX_SOURCE_DIR,
            SPHINX_OUTPUT_DIR,
        )


@nox.session
def docs_autobuild(session):
    session.install(PIP_SPEC)
    session.install("-e", ".[docs]")
    session.install("sphinx_autobuild")
    session.run("python", "-m", "sphinx_autobuild", *SPHINX_HTML_ARGS, "--watch", "src")


@nox.session
@nox.parametrize("python", ["3.7", "3.8", "3.9", "3.10", "3.11"])
def test(session):
    args = Parser().parse_args(session.posargs)

    install_deps = {PYTEST_SPEC, PYTEST_XDIST_SPEC}
    install_extras = set()
    pytest_commands = []

    # Group specific behavior:

    if "unit" in args.test_groups:
        install_deps |= {PYTEST_MOCK_SPEC}
        install_extras |= {"algo"}
        pytest_commands.extend(
            [
                ["-n", "auto", "-p", "no:pytest-qt", "tests/unit"],
                ["-n", "auto", "-p", "no:pytest-qt", "tests/processing"],
            ]
        )

    if "integration" in args.test_groups:
        real_args = [arg for arg in args.integration_args if arg not in {"a111", "a121"}]
        if "a111" in args.integration_args:
            pytest_commands.extend(
                [
                    ["-p", "no:pytest-qt", "tests/integration/a111", *real_args],
                ]
            )
        elif "a121" in args.integration_args:
            pytest_commands.extend(
                [
                    ["-p", "no:pytest-qt", "tests/integration/a121", *real_args],
                ]
            )
        else:
            session.error(
                "The 'integration' session needs to be passed"
                + "either 'a121' or 'a111' in its '--integration-args'"
            )

    if "app" in args.test_groups:
        install_deps |= {
            PYTEST_MOCK_SPEC,
            "pytest-qt",
            "pytest-timeout",
            "requests",
        }
        install_extras |= {"app"}
        pytest_commands.extend(
            [
                ["--timeout=120", "--timeout_method=thread", "tests/gui"],
            ]
        )

    # Override pytest command:

    if args.pytest_args is not None:
        pytest_commands = [args.pytest_args]

    # Install and run:

    install = []

    if args.editable:
        session.install(PIP_SPEC)
        install.append("-e")

    if install_extras:
        install.append(f".[{','.join(install_extras)}]")
    else:
        install.append(".")

    install.extend(install_deps)

    session.install(*install)

    for cmd in pytest_commands:
        session.run("python", "-m", "pytest", *cmd)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from .app_model import PluginSpec


def load_default_plugins() -> list[PluginSpec]:
    from acconeer.exptool.a121.algo.bilateration._plugin import BILATERATION_PLUGIN
    from acconeer.exptool.a121.algo.distance._detector_plugin import DISTANCE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.phase_tracking._plugin import PHASE_TRACKING_PLUGIN
    from acconeer.exptool.a121.algo.presence._detector_plugin import PRESENCE_DETECTOR_PLUGIN
    from acconeer.exptool.a121.algo.sparse_iq._plugin import SPARSE_IQ_PLUGIN
    from acconeer.exptool.a121.algo.touchless_button._plugin import TOUCHLESS_BUTTON_PLUGIN
    from acconeer.exptool.a121.algo.vibration._plugin import VIBRATION_PLUGIN

    return [
        SPARSE_IQ_PLUGIN,
        DISTANCE_DETECTOR_PLUGIN,
        PHASE_TRACKING_PLUGIN,
        PRESENCE_DETECTOR_PLUGIN,
        TOUCHLESS_BUTTON_PLUGIN,
        VIBRATION_PLUGIN,
        BILATERATION_PLUGIN,
    ]



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from __future__ import annotations

import abc
import json
import logging
import operator
import re
import signal
import struct
import sys
import time
from datetime import datetime
from typing import Any, List, Optional

import attrs
import numpy as np
import serial.tools.list_ports
from packaging import version


try:
    from PySide6 import QtCore
except ImportError:
    QtCore = None

try:
    import pyqtgraph as pg
except ImportError:
    pg = None

try:
    from acconeer.exptool._winusbcdc.winusbpy import WinUsbPy
except ImportError:
    WinUsbPy = None

from acconeer.exptool._pyusb.pyusbcomm import PyUsbComm


@attrs.frozen(kw_only=True)
class CommDevice(abc.ABC):
    name: Optional[str] = None
    serial: Optional[str] = None
    unflashed: bool = False
    recognized: bool = False

    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls, d: dict) -> CommDevice:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> CommDevice:
        return cls.from_dict(json.loads(json_str))

    @abc.abstractmethod
    def display_name(self) -> str:
        pass


@attrs.frozen(kw_only=True)
class SerialDevice(CommDevice):
    port: str

    def display_name(self) -> str:
        if self.name is not None:
            dev_name = f"{self.name} {self.port}"
        else:
            dev_name = f"{self.port}"

        if self.unflashed:
            display_name = f"Unflashed {dev_name}"
        elif self.serial is not None:
            display_name = f"{dev_name} ({self.serial})"
        else:
            display_name = dev_name
        return display_name

    def __str__(self) -> str:
        name_str = "" if self.name is None else f"{self.name}: "
        serial_str = "?" if self.serial is None else self.serial
        return f"{name_str}port={self.port} serial={serial_str}"


@attrs.frozen(kw_only=True)
class USBDevice(CommDevice):
    vid: int
    pid: int
    accessible: bool = True

    def display_name(self) -> str:
        if not self.accessible:
            display_name = f"{self.name} (inaccessible)"
        elif self.unflashed:
            display_name = f"Unflashed {self.name}"
        elif self.serial is not None:
            display_name = f"{self.name} ({self.serial})"
        else:
            display_name = f"{self.name}"
        return display_name

    def __str__(self) -> str:
        serial_str = "?" if self.serial is None else self.serial
        return (
            f"{self.name}: USB_VID=0x{self.vid:04x}, USB_PID=0x{self.pid:04x} serial={serial_str}"
        )


_USB_IDS = [  # (vid, pid, 'model number', 'Unflashed')
    (0x0483, 0xA41D, "XC120", True),
    (0x0483, 0xA42C, "XC120", True),
    (0x0483, 0xA42D, "XC120", False),
    (0x0483, 0xA449, "XC120", False),
    (0xACC0, 0xE121, "XV12X", False),
]


class ExampleInterruptHandler:
    def __init__(self):
        self._signal_count = 0
        signal.signal(signal.SIGINT, self.interrupt_handler)

    @property
    def got_signal(self):
        return self._signal_count > 0

    def force_signal_interrupt(self):
        self.interrupt_handler(signal.SIGINT, None)

    def interrupt_handler(self, signum, frame):
        self._signal_count += 1
        if self._signal_count >= 3:
            raise KeyboardInterrupt


def mpl_setup_yaxis_for_phase(ax):
    ax.set_ylim(-np.pi, np.pi)
    ax.set_yticks(np.linspace(-np.pi, np.pi, 5))
    ax.set_yticklabels([r"$-\pi$", r"$-\pi/2$", r"0", r"$\pi/2$", r"$\pi$"])


def timestamp():
    """Get the current date and time formatted as 1999-12-31_23-59-59"""

    return datetime.now().strftime("%Y-%m-%d_%H-%M-%S")


def config_logging(args=None, level=logging.WARN):
    fmt = "{asctime}.{msecs:03.0f} | {levelname:<7} | {processName:<16} | {name:<36} | {message}"
    datefmt = "%H:%M:%S"

    if args is not None:
        if args.debug:
            level = logging.DEBUG
        elif args.verbose:
            level = logging.INFO
        elif args.quiet:
            level = logging.ERROR

    stream_handler = logging.StreamHandler()
    formatter = logging.Formatter(fmt, datefmt=datefmt, style="{")
    stream_handler.setFormatter(formatter)
    log = logging.getLogger(__name__.split(".")[0])
    log.setLevel(level)
    log.addHandler(stream_handler)

    logging.getLogger(__name__).debug("logging configured")


def set_loglevel(level):
    log = logging.getLogger(__name__.split(".")[0])
    log.setLevel(level)


def serial_device_from_port_object(port_object):

    # Return USB serial port if existing in _USB_IDS
    for vid, pid, model_number, unflashed in _USB_IDS:
        if port_object.vid == vid and port_object.pid == pid:
            return SerialDevice(
                name=model_number,
                port=port_object.device,
                serial=port_object.serial_number,
                unflashed=unflashed,
                recognized=True,
            )

    # Find serial portTryReturn USB serial port if existing in _USB_IDS
    device_port = port_object.device
    device_name = None
    device_serial_number = port_object.serial_number
    device_unflashed = False
    device_recognized = False

    PRODUCT_REGEX = r"[X][A-Z]\d{3}"
    desc = port_object.product or port_object.description
    match = re.search(PRODUCT_REGEX, desc)

    if match is None:
        pass
    elif match.group().lower() in ["xe123", "xe124", "xe125", "xe132"]:
        if version.parse(serial.__version__) >= version.parse("3.5"):
            # Special handling of cp2105 modules with with pyserial >= 3.5
            interface = port_object.interface

            if interface and "enhanced" in interface.lower():
                # Add the "enhanced" interface
                device_name = match.group().upper()
                device_recognized = True

        else:  # pyserial <= 3.4
            # Add "?" to both to indicate that it could be either.
            device_name = f"{match.group().upper()} (?)"
    else:
        device_recognized = True
        device_name = f"{match.group().upper()}"

    if desc is not None and "Bootloader" in desc:
        device_unflashed = True

    return SerialDevice(
        name=device_name,
        port=device_port,
        serial=device_serial_number,
        unflashed=device_unflashed,
        recognized=device_recognized,
    )


def get_serial_devices() -> List[SerialDevice]:
    serial_devices = []

    port_objects = serial.tools.list_ports.comports()
    for port_object in port_objects:
        serial_devices.append(serial_device_from_port_object(port_object))

    return serial_devices


def tag_serial_ports_objects(port_infos):
    PRODUCT_REGEX = r"[X][A-Z]\d{3}"

    port_tag_tuples = []

    for port_object in port_infos:
        desc = port_object.product or port_object.description

        match = re.search(PRODUCT_REGEX, desc)
        if match is None:
            for vid, pid, model_number, _ in _USB_IDS:
                if port_object.vid == vid and port_object.pid == pid:
                    port_tag_tuples.append((port_object, model_number))
                    break
            else:
                port_tag_tuples.append((port_object, None))
        elif match.group().lower() in ["xe123", "xe124", "xe125", "xe132"]:
            if version.parse(serial.__version__) >= version.parse("3.5"):
                # Special handling of cp2105 modules with with pyserial >= 3.5
                interface = port_object.interface

                if interface and "enhanced" in interface.lower():
                    # Add the "enhanced" interface
                    port_tag_tuples.append((port_object, f"{match.group().upper()}"))
                else:
                    # Add the "standard" interface but don't tag it
                    port_tag_tuples.append((port_object, None))

            else:  # pyserial <= 3.4
                # Add "?" to both to indicate that it could be either.
                port_tag_tuples.append((port_object, f"{match.group().upper()} (?)"))
        elif "Bootloader" in desc:
            port_tag_tuples.append((port_object, f"Unflashed {match.group()}"))
        else:
            port_tag_tuples.append((port_object, f"{match.group()}"))

    return port_tag_tuples


def tag_serial_ports(port_infos):
    """
    Returns every port and Acconeer model number in a tuple if it's an
    Acconeer product.

    E.g:
    [ ("/dev/ttyUSB0", None),   <- Not an Acconeer product
      ("/dev/ttyUSB1", "XB112") <- An Acconeer product
    ]

    :param port_infos: Information about ports as given by
        `serial.tools.list_ports.comports()`
    :returns: List of tuples [(<port>, <model number> or None), ...]
    """
    port_tag_tuples = tag_serial_ports_objects(port_infos)
    port_tag_tuples = [(port.device, tag) for (port, tag) in port_tag_tuples]
    return port_tag_tuples


def get_tagged_serial_ports():
    tagged_serial_ports = []
    serial_devices = get_serial_devices()
    for serial_device in serial_devices:
        tagged_serial_ports.append((serial_device.port, serial_device.name))
    return tagged_serial_ports


def autodetect_serial_port():
    port_infos = serial.tools.list_ports.comports()

    tagged_serial_ports = tag_serial_ports(port_infos)
    acconeer_port_infos = [pinfo for pinfo in tagged_serial_ports if pinfo[1]]

    # Check for mutliple Acconeer devices;
    if len(acconeer_port_infos) > 1:
        print("Found multiple Acconeer products:", end="\n\n")
        for port, tag in [("Serial port:", "Model:")] + acconeer_port_infos:
            print(f"\t{port:<15} {tag:}")
        print('\nRun the script again and specify port using the "-u"/"--uart" flag.')
        sys.exit()

    if len(acconeer_port_infos) == 1:
        port, tag = acconeer_port_infos[0]
        print("Autodetected {} on {}\n".format(tag, port))
        return port

    for port_info in port_infos:
        port, desc, _ = port_info
        if desc == "FT230X Basic UART":
            print("Autodetected FT230X Basic UART on {}\n".format(port))
            return port

    if len(port_infos) == 0:
        print("Could not autodetect serial port, no ports available")
        sys.exit()
    elif len(port_infos) == 1:
        print("Autodetected single available serial port on {}\n".format(port))
        return port_infos[0][0]
    else:
        print("Multiple serial ports are available:")
        for port_info in port_infos:
            port, desc, _ = port_info
            print("  {:<13}  ({})".format(port, desc))
        print("\nRe-run the script with a given port")
        sys.exit()

    print("Could not autodetect serial port")
    sys.exit()


def get_usb_devices(only_accessible=False):
    usb_devices = []

    if WinUsbPy is not None:
        winusbpy = WinUsbPy()
        device_list = winusbpy.find_all_devices()
        for device in device_list:
            device_vid = device["vid"]
            device_pid = device["pid"]
            serial_number = device["serial"]
            for vid, pid, model_name, unflashed in _USB_IDS:
                if device_vid == vid and device_pid == pid:
                    usb_devices.append(
                        USBDevice(
                            vid=device_vid,
                            pid=device_pid,
                            serial=serial_number,
                            name=model_name,
                            unflashed=unflashed,
                            recognized=True,
                        )
                    )
    else:
        pyusbcomm = PyUsbComm()
        for device_vid, device_pid, serial_number in pyusbcomm.iterate_devices():
            for vid, pid, model_name, unflashed in _USB_IDS:
                if device_vid == vid and device_pid == pid:
                    device_name = model_name
                    accessible = pyusbcomm.is_accessible(vid, pid)
                    if only_accessible and not accessible:
                        continue

                    usb_devices.append(
                        USBDevice(
                            vid=device_vid,
                            pid=device_pid,
                            serial=serial_number,
                            name=device_name,
                            accessible=accessible,
                            unflashed=unflashed,
                            recognized=True,
                        )
                    )

    return usb_devices


def get_usb_device_by_serial(serial, only_accessible=False):
    usb_devices = get_usb_devices(only_accessible=only_accessible)
    if serial is not None:
        for device in usb_devices:
            if serial == device.serial:
                return device
    raise ValueError(f"Could not find usb device with serial number '{serial}'")


def color_cycler(i=0):
    category10 = [
        "#1f77b4",
        "#ff7f0e",
        "#2ca02c",
        "#d62728",
        "#9467bd",
        "#8c564b",
        "#e377c2",
        "#7f7f7f",
        "#bcbd22",
        "#17becf",
    ]

    return category10[i % len(category10)]


def pg_pen_cycler(i=0, style=None, width=2):
    pen = pg.mkPen(color_cycler(i), width=width)
    if style == "--":
        pen.setStyle(QtCore.Qt.DashLine)
    elif style is not None:
        pen.setStyle(style)
    return pen


def pg_brush_cycler(i=0):
    return pg.mkBrush(color_cycler(i))


class SmoothMax:
    def __init__(self, f=None, hysteresis=0.5, tau_decay=2.0, tau_grow=0.5):
        self.fixed_dt = 1 / f if (f is not None and f > 0) else None
        self.hyst = hysteresis
        self.tc_decay = tau_decay
        self.tc_grow = tau_grow

        self.last_t = 0
        self.x = self.y = -1

    def update(self, data):
        m = max(np.nanmax(data), 1e-12)

        if self.fixed_dt is None:
            now = time.perf_counter()
            dt = now - self.last_t
            self.last_t = now
        else:
            dt = self.fixed_dt

        ax = np.exp(-dt / self.tc_decay) if self.tc_decay > 1e-3 else 0
        ay = np.exp(-dt / self.tc_grow) if self.tc_grow > 1e-3 else 0

        if m > self.x:
            self.x = m
        else:
            self.x = (1 - ax) * m + ax * self.x

        if self.y < 0:
            self.y = self.x
        elif self.x > self.y:
            self.y = (1 - ay) * self.x * (1 + self.hyst) + ay * self.y
        elif self.x < (1 - self.hyst) * self.y:
            self.y = self.x / (1 - self.hyst)

        return self.y


class SmoothLimits:
    def __init__(self, f=None, hysteresis=0.3, tau_decay=1.5, tau_grow=0.3):
        self.fixed_dt = 1 / f if (f is not None and f > 0) else None
        self.hyst = hysteresis
        self.tc_decay = tau_decay
        self.tc_grow = tau_grow

        self.last_t = 0
        self.x = self.y = self.z = None

    def update(self, data):
        data_lims = (np.nanmin(data), np.nanmax(data))

        if self.fixed_dt is None:
            now = time.perf_counter()
            dt = now - self.last_t
            self.last_t = now
        else:
            dt = self.fixed_dt

        if self.x is None:  # First call
            self.x = list(data_lims)
            self.y = list(data_lims)
            self.z = list(data_lims)
            return list(data_lims)

        ad = np.exp(-dt / self.tc_decay) if self.tc_decay > 1e-3 else 0
        ag = np.exp(-dt / self.tc_grow) if self.tc_grow > 1e-3 else 0

        ops = (operator.lt, operator.gt)
        idxs = (0, 1)

        for op, i in zip(ops, idxs):
            if op(data_lims[i], self.x[i]):
                self.x[i] = data_lims[i]
            else:
                self.x[i] = (1 - ad) * data_lims[i] + ad * self.x[i]

        for op, i in zip(ops, idxs):
            if op(self.x[i], self.y[i]):
                self.y[i] = (1 - ag) * self.x[i] + ag * self.y[i]
            else:
                self.y[i] = self.x[i]

        for op, i, j in zip(ops, idxs, reversed(idxs)):
            y_hyst = self.y[i] + (self.y[i] - self.y[j]) * (1 / (1 - self.hyst) - 1)

            if op(self.y[i], self.z[i]):
                self.z[i] = self.y[i]
            elif op(self.z[i], y_hyst):
                self.z[i] = y_hyst

        return self.z


pg_phase_ticks = [
    list(zip(np.linspace(-np.pi, np.pi, 5), ["-π", "-π/2", "0", "π/2", "π"])),
    [(x, "") for x in np.linspace(-np.pi, np.pi, 9)],
]


def pg_setup_polar_plot(plot, max_r=1):
    plot.showAxis("left", False)
    plot.showAxis("bottom", False)
    plot.setAspectLocked()
    plot.disableAutoRange()
    s = 1.15
    plot.setXRange(-s * max_r, s * max_r)
    plot.setYRange(-s * max_r, s * max_r)

    for i, r in enumerate(np.linspace(0, max_r, 5)[1:]):
        circle = pg.QtWidgets.QGraphicsEllipseItem(-r, -r, r * 2, r * 2)
        circle.setPen(pg.mkPen("k" if i == 3 else 0.5))
        plot.addItem(circle)

        if i == 3:
            text_item = pg.TextItem(text=str(r), color="k", anchor=(0, 1))
            text_item.setPos(np.cos(np.pi / 8) * r, np.sin(np.pi / 8) * r)
            plot.addItem(text_item)

    for i in range(8):
        deg = (360 / 8) * i
        rad = np.radians(deg)
        x = np.cos(rad)
        y = np.sin(rad)
        text = str(int(deg)) + "\u00b0"
        ax = (-x + 1) / 2
        ay = (y + 1) / 2
        text_item = pg.TextItem(text, color="k", anchor=(ax, ay))
        text_item.setPos(max_r * x * 1.02, max_r * y * 1.02)
        plot.addItem(text_item)
        plot.plot([0, max_r * x], [0, max_r * y], pen=pg.mkPen(0.5))


def pg_mpl_cmap(name):
    import matplotlib.pyplot as plt

    cmap = plt.get_cmap(name)
    cmap._init()
    return (np.array(cmap._lut) * 255).astype(np.uint8)


class FreqCounter:
    """
    Helper class for measuring & calculating temporal information:
        * time between ticks (lp-filtered), `FreqCounter.lp_dt`
        * frequency
        * data throughput (if num_bits is not None)
    """

    def __init__(self, a=None, tc=None, num_bits=None):

        assert (a is None) or (tc is None)

        if (a is None) and (tc is None):
            tc = 1.0

        self.a = a
        self.tc = tc
        self.num_bits = num_bits
        self.avg_dt_buf_len = 150  # Chosen after empirical testing
        self.reset()

    def reset(self):
        self.last_t = None
        self.lp_avg_dt = None
        self.num_ticks = 0
        self.avg_dt_buf = np.full(self.avg_dt_buf_len, np.nan)

    def tick_values(self):
        """
        Ticks the FreqCounter by taking the time delta from last time this was called.
        The update is done by a running average over the last 5 sweeps.
        The averaged value is passed through a lp_filter to avoid displaying blinking numbers.

        OBS! The first call to this method will return None.

        :returns:
            3-tuple of (delta time (averaged and lp-filtered), frequency, [throughput])
            -- OR --
            None
        """

        now = time.perf_counter()

        if self.last_t is None:
            self.last_t = now
            return None

        dt = now - self.last_t

        self.avg_dt_buf = np.roll(self.avg_dt_buf, -1)
        self.avg_dt_buf[-1] = dt

        avg_dt = np.nanmean(self.avg_dt_buf)

        if self.a is not None:
            a = self.a
        else:
            a = np.exp(-avg_dt / self.tc)

        a = min(a, 1.0 - 1.0 / (1.0 + self.num_ticks))

        if self.lp_avg_dt is None:
            self.lp_avg_dt = avg_dt
        else:
            self.lp_avg_dt = a * self.lp_avg_dt + (1 - a) * avg_dt

        self.last_t = now
        self.num_ticks += 1

        f = 1 / self.lp_avg_dt

        if self.num_bits is None:
            data_rate = None
        else:
            data_rate = self.num_bits * f

        return self.lp_avg_dt, f, data_rate

    def tick(self):
        """
        Prints the values returned by `tick_values()`.
        """
        tick_info = self.tick_values()
        if tick_info is None:
            return

        dt, f, data_rate = tick_info
        dt_ms = dt * 1e3
        data_rate_mbps = data_rate * 1e-6

        if data_rate is None:
            print(" {:5.1f} ms, {:5.1f} Hz".format(dt_ms, f), end="\r")
        else:
            s = " {:5.1f} ms, {:5.1f} Hz, {:5.2f} Mbit/s".format(dt_ms, f, data_rate_mbps)
            print(s, end="\r")


def hex_to_rgb_tuple(hex_color):
    return struct.unpack("BBB", bytes.fromhex(hex_color.lstrip("#")))


def is_power_of_2(n):
    return (n & (n - 1) == 0) and n != 0


def optional_or_else(value, default):
    return default if value is None else value



# Copyright (c) Acconeer AB, 2022-2023
# All rights reserved

from acconeer.exptool.flash._stm32uart import ACCONEER_XM_CP2105_MODULE_PID, Stm32UartFlasher
from acconeer.exptool.flash._xc120 import (
    ACCONEER_XC120_BOOTLOADER_PID,
    ACCONEER_XC120_EXPLORATION_SERVER_PID,
    ACCONEER_XC120_EXPLORATION_SERVER_WINUSB_PID,
    BootloaderTool,
)


PRODUCT_PID_TO_FLASH_MAP = {
    ACCONEER_XC120_BOOTLOADER_PID: BootloaderTool,
    ACCONEER_XC120_EXPLORATION_SERVER_PID: BootloaderTool,
    ACCONEER_XC120_EXPLORATION_SERVER_WINUSB_PID: BootloaderTool,
    ACCONEER_XM_CP2105_MODULE_PID: Stm32UartFlasher,
}

PRODUCT_NAME_TO_FLASH_MAP = {
    "XC120": BootloaderTool,
    "XE125": Stm32UartFlasher,
    "XM125": Stm32UartFlasher,
}

EVK_TO_PRODUCT_MAP = {
    "XC120": "XC120",
    "XE125": "XM125",
    "XM125": "XM125",
}



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from pathlib import Path
from typing import Optional

import attrs
import h5py

from acconeer.exptool.app.new._enums import PluginGeneration


@attrs.frozen(kw_only=True)
class FileFindings:
    generation: PluginGeneration
    key: Optional[str]


def investigate_file(path: Path) -> Optional[FileFindings]:
    if path.suffix != ".h5":
        return None

    try:
        f = h5py.File(path, "r")
    except Exception:
        return None

    try:
        try:
            generation = PluginGeneration(bytes(f["generation"][()]).decode())
        except Exception:
            return None

        try:
            key = bytes(f["algo"]["key"][()]).decode()
        except KeyError:
            key = None
    finally:
        f.close()

    if generation != PluginGeneration.A121:
        return None

    return FileFindings(
        generation=generation,
        key=key,
    )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import logging
import multiprocessing as mp
import queue
import time
import traceback
import uuid
from multiprocessing.synchronize import Event as mp_EventType  # NOTE! this is not mp.Event.
from typing import Any, Dict, Optional, Tuple, Union

import attrs
import psutil
from typing_extensions import Literal

from ._backend_logger import BackendLogger
from ._message import GeneralMessage, Message
from ._model import Model


log = logging.getLogger(__name__)


@attrs.frozen
class ClosedTask:
    key: uuid.UUID = attrs.field()
    exception: Optional[Exception] = attrs.field(default=None)
    traceback_format_exc: Optional[str] = attrs.field(default=None)


TaskName = str
TaskPlugin = bool
TaskKwargs = Dict[str, Any]
Task = Tuple[TaskName, TaskKwargs, TaskPlugin]

ToBackendQueueItem = Union[
    Tuple[Literal["stop"], None],
    Tuple[Literal["task"], Tuple[uuid.UUID, Task]],
]
FromBackendQueueItem = Union[Message, ClosedTask]


class Backend:
    def __init__(self) -> None:
        self._recv_queue: mp.Queue[FromBackendQueueItem] = mp.Queue()
        self._send_queue: mp.Queue[ToBackendQueueItem] = mp.Queue()
        self._stop_event = mp.Event()
        self._process = mp.Process(
            target=process_program,
            args=(
                self._send_queue,
                self._recv_queue,
                self._stop_event,
            ),
            daemon=True,
        )

    def start(self) -> None:
        log.debug("Backend starting ...")
        self._process.start()

    def stop(self) -> None:
        log.debug("Backend stopping ...")
        self._stop_event.set()
        self._send(("stop", None))

        self._process.join(timeout=3)

        if self._process.exitcode is None:
            log.warning("Backend process join timed out, killing...")

            self._process.kill()
            self._process.join(timeout=1)

        if self._process.exitcode is None:
            raise RuntimeError

        self._process.close()

    def put_task(self, task: Task) -> uuid.UUID:
        key = uuid.uuid4()
        self._send(("task", (key, task)))
        return key

    def _send(self, item: ToBackendQueueItem) -> None:
        self._send_queue.put(item)

    def recv(self, timeout: Optional[float] = None) -> FromBackendQueueItem:
        return self._recv_queue.get(timeout=timeout)


def process_program(
    recv_queue: mp.Queue[ToBackendQueueItem],
    send_queue: mp.Queue[FromBackendQueueItem],
    stop_event: mp_EventType,
) -> None:
    MAX_POLL_INTERVAL = 0.5

    process = psutil.Process()
    process.cpu_percent()
    last_cpu_msg_time = time.monotonic()

    try:
        BackendLogger.set_callback(send_queue.put)
        process_log = BackendLogger.getLogger(__name__)
        model = Model(task_callback=send_queue.put)
        model_wants_to_idle = False

        while not stop_event.is_set():
            now = time.monotonic()
            if now - last_cpu_msg_time > MAX_POLL_INTERVAL:
                last_cpu_msg_time = now
                cpu_percent = round(process.cpu_percent())
                send_queue.put(
                    GeneralMessage(
                        name="cpu_percent",
                        data=cpu_percent,
                    )
                )

            msg = None

            if not model_wants_to_idle:
                try:
                    msg = recv_queue.get(timeout=MAX_POLL_INTERVAL)
                except queue.Empty:
                    continue

                process_log.debug(f"Backend received the command: {msg}")
            else:
                try:
                    msg = recv_queue.get_nowait()
                    process_log.debug(f"Backend received the command: {msg}")
                except queue.Empty:
                    pass

            if msg is None:  # Model wanted idle and nothing in queue
                try:
                    model_wants_to_idle = model.idle()
                except Exception as exc:
                    model_wants_to_idle = False
                    send_queue.put(
                        GeneralMessage(
                            name="error",
                            exception=exc,
                            traceback_format_exc=traceback.format_exc(),
                        )
                    )

                continue

            cmd, maybe_key_and_task = msg

            if cmd == "stop":
                break
            elif cmd == "task":
                assert maybe_key_and_task is not None

                key, task = maybe_key_and_task
                name, kwargs, plugin = task

                try:
                    model.execute_task(name, kwargs, plugin)
                except Exception as exc:
                    send_queue.put(ClosedTask(key, exc, traceback.format_exc()))
                else:
                    send_queue.put(ClosedTask(key))

                model_wants_to_idle = True
            else:
                raise RuntimeError
    finally:
        recv_queue.close()
        send_queue.close()



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from .backend_plugin import (
    GenericProcessorBackendPluginBase,
    ProcessorBackendPluginBase,
    ProcessorBackendPluginSharedState,
    ProcessorPluginPreset,
)
from .plot_plugin import (
    ExtendedProcessorPlotPluginBase,
    GenericProcessorPlotPluginBase,
    ProcessorPlotPluginBase,
)
from .plugin import ProcessorPluginSpec
from .view_plugin import ProcessorViewPluginBase



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

import attrs
import numpy as np
import numpy.typing as npt

from acconeer.exptool import a121
from acconeer.exptool.a121.algo import (
    APPROX_BASE_STEP_LENGTH_M,
    AlgoProcessorConfigBase,
    ProcessorBase,
)


@attrs.mutable(kw_only=True)
class ProcessorConfig(AlgoProcessorConfigBase):
    time_series_length: int = attrs.field(default=512)
    """Length of time series."""

    lp_coeff: float = attrs.field(default=0.75)
    """Specify filter coefficient of exponential filter."""

    sensitivity: float = attrs.field(default=10.0)
    """Specify threshold sensitivity."""

    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        validation_results: list[a121.ValidationResult] = []

        if config.sensor_config.sweep_rate is None:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "sweep_rate",
                    "Must be set",
                )
            )

        if config.sensor_config.num_points != 1:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config.subsweep,
                    "num_points",
                    "Must be set to 1",
                )
            )

        if not config.sensor_config.double_buffering:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "double_buffering",
                    "Must be enabled",
                )
            )

        if not config.sensor_config.continuous_sweep_mode:
            validation_results.append(
                a121.ValidationError(
                    config.sensor_config,
                    "continuous_sweep_mode",
                    "Must be enabled",
                )
            )

        return validation_results


@attrs.frozen(kw_only=True)
class ProcessorContext:
    ...


@attrs.frozen(kw_only=True)
class ProcessorResult:
    time_series: npt.NDArray[np.float_]
    lp_z_abs_db: npt.NDArray[np.float_]
    freqs: npt.NDArray[np.float_]
    max_amplitude: float
    max_psd_ampl: float
    max_psd_ampl_freq: float


class Processor(ProcessorBase[ProcessorConfig, ProcessorResult]):

    _OVER_SAMPLING_FACTOR = 2
    _WINDOW_BASE_LENGTH = 10
    _HALF_GUARD_BASE_LENGTH = 5

    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: a121.Metadata,
        processor_config: ProcessorConfig,
        subsweep_indexes: Optional[list[int]] = None,
        context: Optional[ProcessorContext] = None,
    ) -> None:

        processor_config.validate(sensor_config)

        # Should never happen, checked in validate
        assert sensor_config.sweep_rate is not None

        self.spf = sensor_config.sweeps_per_frame
        self.lp_coeffs = processor_config.lp_coeff
        self.sensitivity = processor_config.sensitivity
        self.time_series_length = processor_config.time_series_length

        self.time_series = np.zeros(shape=processor_config.time_series_length)
        self.freq = np.fft.rfftfreq(
            self._OVER_SAMPLING_FACTOR * processor_config.time_series_length,
            1 / sensor_config.sweep_rate,
        )[1:]
        self.lp_z_abs_db = np.zeros_like(self.freq)
        self.window_length = self._WINDOW_BASE_LENGTH * self._OVER_SAMPLING_FACTOR
        self.half_guard_length = self._HALF_GUARD_BASE_LENGTH * self._OVER_SAMPLING_FACTOR

    def process(self, result: a121.Result) -> ProcessorResult:

        new_data_segment = np.angle(result.frame.squeeze(axis=1))

        self.time_series = np.roll(self.time_series, -self.spf)
        self.time_series[-self.spf :] = new_data_segment
        self.time_series = np.unwrap(self.time_series)

        z_abs = np.abs(
            np.fft.rfft(
                self.time_series - np.mean(self.time_series),
                n=self.time_series_length * self._OVER_SAMPLING_FACTOR,
            )
        )[1:]
        z_abs_db = 20 * np.log10(z_abs)
        self.lp_z_abs_db = self.lp_z_abs_db * self.lp_coeffs + z_abs_db * (1 - self.lp_coeffs)

        presented_time_series = (
            (self.time_series - np.mean(self.time_series)) * APPROX_BASE_STEP_LENGTH_M * 1000
        )

        max_amplitude = float(np.max(np.abs(result.frame)))

        threshold = self._calculate_cfar_threshold(
            self.lp_z_abs_db, self.sensitivity, self.window_length, self.half_guard_length
        )
        # Find index of values over threshold
        idx_over_threshold = np.where(threshold < self.lp_z_abs_db)[0]
        if idx_over_threshold.shape[0] != 0:
            psd_ampls_over_threshold = self.lp_z_abs_db[idx_over_threshold]
            max_psd_ampl = np.max(psd_ampls_over_threshold)
            max_psd_ampl_freq = self.freq[idx_over_threshold[np.argmax(psd_ampls_over_threshold)]]
        else:
            max_psd_ampl = None
            max_psd_ampl_freq = None

        return ProcessorResult(
            time_series=presented_time_series,
            lp_z_abs_db=self.lp_z_abs_db,
            freqs=self.freq,
            max_amplitude=max_amplitude,
            max_psd_ampl=max_psd_ampl,
            max_psd_ampl_freq=max_psd_ampl_freq,
        )

    @staticmethod
    def _calculate_cfar_threshold(
        psd: npt.NDArray[np.float_],
        sensitivity: float,
        window_length: int,
        half_guard_length: int,
    ) -> npt.NDArray[np.float_]:
        threshold = np.full(psd.shape, np.nan)
        margin = window_length + half_guard_length
        length_after_filtering = psd.shape[0] - 2 * margin

        filt_psd = np.convolve(psd, np.ones(window_length), "valid") / window_length
        threshold[margin:-margin] = (
            filt_psd[:length_after_filtering] + filt_psd[-length_after_filtering:]
        ) / 2 + sensitivity

        return threshold

    def update_config(self, config: ProcessorConfig) -> None:
        ...


def get_sensor_config() -> a121.SensorConfig:
    return a121.SensorConfig(
        profile=a121.Profile.PROFILE_3,
        hwaas=16,
        num_points=1,
        step_length=1,
        start_point=80,
        receiver_gain=10,
        sweep_rate=2000,
        sweeps_per_frame=50,
        double_buffering=True,
        continuous_sweep_mode=True,
        inter_frame_idle_state=a121.IdleState.READY,
        inter_sweep_idle_state=a121.IdleState.READY,
    )



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import abc
import enum
import json
import warnings
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

import attrs

from acconeer.exptool import a121
from acconeer.exptool.a121._core.utils import EntityJSONEncoder


AlgoBaseT = TypeVar("AlgoBaseT", bound="AlgoBase")
InputT = TypeVar("InputT", a121.Result, List[Dict[int, a121.Result]])
MetadataT = TypeVar("MetadataT", a121.Metadata, List[Dict[int, a121.Metadata]])
ConfigT = TypeVar("ConfigT", bound="AlgoConfigBase")
ProcessorConfigT = TypeVar("ProcessorConfigT", bound="AlgoProcessorConfigBase")
ParamEnumT = TypeVar("ParamEnumT", bound="AlgoParamEnum")
ResultT = TypeVar("ResultT")


class GenericProcessorBase(abc.ABC, Generic[InputT, ProcessorConfigT, ResultT, MetadataT]):
    def __init__(
        self,
        *,
        sensor_config: a121.SensorConfig,
        metadata: MetadataT,
        processor_config: ProcessorConfigT,
    ) -> None:
        self.sensor_config = sensor_config
        self.metadata: MetadataT = metadata
        self.processor_config = processor_config

    @abc.abstractmethod
    def process(self, result: InputT) -> ResultT:
        ...

    @abc.abstractmethod
    def update_config(self, config: ProcessorConfigT) -> None:
        ...


ProcessorBase = GenericProcessorBase[a121.Result, ProcessorConfigT, ResultT, a121.Metadata]
ExtendedProcessorBase = GenericProcessorBase[
    List[Dict[int, a121.Result]], ProcessorConfigT, ResultT, List[Dict[int, a121.Metadata]]
]


class AlgoBase:
    def to_dict(self) -> dict[str, Any]:
        return attrs.asdict(self)

    @classmethod
    def from_dict(cls: type[AlgoBaseT], d: dict[str, Any]) -> AlgoBaseT:
        return cls(**d)

    def to_json(self) -> str:
        return json.dumps(self.to_dict(), cls=EntityJSONEncoder)

    @classmethod
    def from_json(cls: type[AlgoBaseT], json_str: str) -> AlgoBaseT:
        return cls.from_dict(json.loads(json_str))


class AlgoConfigBase(AlgoBase):
    def validate(self) -> None:
        """Performs self-validation

        :raises ValidationError: If anything is invalid.
        """
        for validation_result in self._collect_validation_results():
            try:
                raise validation_result
            except a121.ValidationWarning as vw:
                warnings.warn(vw.message)

    @abc.abstractmethod
    def _collect_validation_results(self) -> list[a121.ValidationResult]:
        pass


class AlgoProcessorConfigBase(AlgoBase):
    def validate(self, config: Union[a121.SensorConfig, a121.SessionConfig]) -> None:
        """Performs self-validation and validation of its session config

        :raises ValidationError: If anything is invalid.
        """
        if isinstance(config, a121.SensorConfig):
            config = a121.SessionConfig(config)

        for validation_result in self._collect_validation_results(config):
            try:
                raise validation_result
            except a121.ValidationWarning as vw:
                warnings.warn(vw.message)

    @abc.abstractmethod
    def _collect_validation_results(
        self, config: a121.SessionConfig
    ) -> list[a121.ValidationResult]:
        pass


class AlgoParamEnum(enum.Enum):
    # TODO: Share with config_enums.py (?)

    @classmethod
    def _missing_(cls: type[ParamEnumT], value: object) -> Optional[ParamEnumT]:
        for member in cls:
            if member.name == value:
                return member

        return None



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

import importlib
import logging
import os
import traceback
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Optional, Tuple

import qtawesome as qta
from requests import Response, Session
from requests.cookies import RequestsCookieJar

from PySide6.QtCore import Qt, QThread, Signal
from PySide6.QtGui import QCloseEvent, QMovie, QTextCursor
from PySide6.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QFileDialog,
    QGridLayout,
    QLabel,
    QLineEdit,
    QPushButton,
    QSizePolicy,
    QStackedWidget,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from acconeer.exptool.app import resources  # type: ignore[attr-defined]
from acconeer.exptool.app.new._enums import ConnectionInterface, ConnectionState
from acconeer.exptool.app.new._exceptions import HandledException
from acconeer.exptool.app.new.app_model import AppModel
from acconeer.exptool.app.new.ui.misc import ExceptionWidget
from acconeer.exptool.flash import (  # type: ignore[import]
    BIN_FETCH_PROMPT,
    ET_DIR,
    DevLicense,
    clear_cookies,
    download,
    flash_image,
    get_content,
    get_cookies,
    get_flash_download_name,
    get_flash_known_devices,
    login,
    save_cookies,
)
from acconeer.exptool.utils import CommDevice  # type: ignore[import]

from .misc import BUTTON_ICON_COLOR, SerialPortComboBox, USBDeviceComboBox


_WRONG_CREDENTIALS_MSG = "<font color='red'>Incorrect username (email) or password</font>"

_LOGGED_IN_MSG = (
    "Logged in to <a href='https://developer.acconeer.com/'>developer.acconeer.com</a>"
)

log = logging.getLogger(__name__)


class _FlashThread(QThread):
    flash_failed = Signal(Exception, str)
    flash_done = Signal()
    flash_progress = Signal(int)

    def __init__(
        self, bin_file: str, flash_device: CommDevice, device_name: Optional[str]
    ) -> None:
        super().__init__()
        self.bin_file = bin_file
        self.flash_device = flash_device
        self.device_name = device_name

    def run(self) -> None:
        def progress_callback(progress: int, end: bool = False) -> None:
            if end:
                self.flash_progress.emit(100)
            else:
                self.flash_progress.emit(progress)

        try:
            flash_image(
                self.bin_file,
                self.flash_device,
                device_name=self.device_name,
                progress_callback=progress_callback,
            )
            self.flash_done.emit()
        except Exception as e:
            log.error(str(e))
            self.flash_failed.emit(HandledException(e), traceback.format_exc())


class _FlashDialog(QDialog):
    flash_done = Signal()

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setWindowTitle("Flash tool")
        self.setMinimumWidth(250)

        vbox = QVBoxLayout(self)
        vbox.setAlignment(Qt.AlignCenter)

        self.loading = QLabel()
        self.loading.setAlignment(Qt.AlignCenter)

        loader_gif = None
        with importlib.resources.path(resources, "loader.gif") as path:
            loader_gif = path

        self.flash_movie = QMovie(str(loader_gif))
        self.loading.setMovie(self.flash_movie)
        vbox.addWidget(self.loading)

        self.flash_label = QLabel(self)
        self.flash_label.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.MinimumExpanding)
        self.flash_label.setAlignment(Qt.AlignCenter)
        vbox.addWidget(self.flash_label)

        self.ok_button = QPushButton(self)
        self.ok_button.setText("OK")
        self.ok_button.setHidden(True)
        self.ok_button.clicked.connect(self._on_ok_click)
        vbox.addWidget(self.ok_button)
        self.setLayout(vbox)

    def flash(self, bin_file: str, flash_device: CommDevice, device_name: Optional[str]) -> None:
        self.flash_thread = _FlashThread(bin_file, flash_device, device_name)
        self.flash_thread.started.connect(self._flash_start)
        self.flash_thread.finished.connect(self.flash_thread.deleteLater)
        self.flash_thread.finished.connect(self._flash_stop)
        self.flash_thread.flash_done.connect(self._flash_done)
        self.flash_thread.flash_failed.connect(self._flash_failed)
        self.flash_thread.flash_progress.connect(self._flash_progress)

        self.flash_thread.start()
        self._flashing = True
        self.exec()

    def _flash_start(self) -> None:
        self.flash_label.setText("Flashing...")
        self.loading.show()
        self.ok_button.setHidden(True)
        self.flash_movie.start()

    def _flash_stop(self) -> None:
        self.flash_movie.stop()
        self.loading.hide()
        self._flashing = False

    def _flash_progress(self, progress: int) -> None:
        if self._flashing:
            self.flash_label.setText(f"Flashing... {progress}%")

    def _flash_done(self) -> None:
        self.flash_done.emit()
        self.flash_label.setText("Flashing done!")
        self.ok_button.setHidden(False)

    def _flash_failed(self, exception: Exception, traceback_str: Optional[str]) -> None:
        self.flash_done.emit()
        self.flash_label.setText("Flashing failed!")
        ExceptionWidget(self, exc=exception, traceback_str=traceback_str).exec()
        self.ok_button.setHidden(False)

    def _on_ok_click(self) -> None:
        self.close()

    def closeEvent(self, event: QCloseEvent) -> None:
        if self._flashing:
            self.flash_thread.terminate()
            self.flash_thread.wait()
        super().closeEvent(event)


class _FlashPopup(QDialog):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.setWindowTitle("Flash tool")
        self.setMinimumWidth(550)

        self.bin_file: Optional[str] = None

        self.authenticating = False
        self.downloading_firmware = False
        self.dev_license = DevLicense()

        browse_button = QPushButton("Browse", self)
        browse_button.clicked.connect(self._browse_file)

        self.file_label = QLineEdit(self)
        self.file_label.setReadOnly(True)
        self.file_label.setEnabled(False)
        self.file_label.setPlaceholderText("<Select a bin file>")

        self.get_latest_button = QPushButton("Get latest bin file", self)
        self.get_latest_button.clicked.connect(self._get_latest_bin_file)

        self.downloaded_version_label = QLineEdit(self)
        self.downloaded_version_label.setReadOnly(True)
        self._reset_download_version()

        self.download_spinner = QLabel()
        self.download_spinner.setHidden(True)

        loader_gif = None
        with importlib.resources.path(resources, "loader.gif") as path:
            loader_gif = path

        self.download_movie = QMovie(str(loader_gif))
        self.download_spinner.setMovie(self.download_movie)

        self.download_status_label = QLabel()
        self.download_status_label.setHidden(True)

        self.interface_dd = QComboBox(self)
        self.interface_dd.addItem("Serial", userData=ConnectionInterface.SERIAL)
        self.interface_dd.addItem("USB", userData=ConnectionInterface.USB)

        self.interface_dd.currentIndexChanged.connect(self._on_interface_dd_change)

        self.device_selection = QComboBox(self)
        self.device_selection.setEnabled(False)
        self.device_name_selection = None
        self.device_selection.currentIndexChanged.connect(self._on_device_selection_change)

        self.stacked = QStackedWidget(self)
        self.stacked.setStyleSheet("QStackedWidget {background-color: transparent;}")
        self.stacked.addWidget(SerialPortComboBox(app_model, self.stacked))
        self.stacked.addWidget(USBDeviceComboBox(app_model, self.stacked))

        self.flash_button = QPushButton("Flash", self)
        self.flash_button.clicked.connect(self._flash)
        self.flash_button.setEnabled(False)

        self.logged_in_label = QLabel(self)
        self.logged_in_label.setOpenExternalLinks(True)

        self.logout_button = QPushButton("Logout", self)
        self.logout_button.clicked.connect(self._logout)
        self._hide_login_status()

        layout = QGridLayout(self)

        # fmt: off
        # Grid layout:                                  row, col, rspan, cspan
        layout.addWidget(browse_button,                 0,   0,   1,     3)    # noqa: E241
        layout.addWidget(self.file_label,               0,   3,   1,     9)    # noqa: E241
        layout.addWidget(self.get_latest_button,        1,   0,   1,     3)    # noqa: E241
        layout.addWidget(self.downloaded_version_label, 1,   3,   1,     9)    # noqa: E241
        layout.addWidget(self.interface_dd,             2,   0,   1,     3)    # noqa: E241
        layout.addWidget(self.stacked,                  2,   3,   1,     6)    # noqa: E241
        layout.addWidget(self.device_selection,         2,   9,   1,     3)    # noqa: E241
        layout.addWidget(self.flash_button,             3,   0,   1,    12)    # noqa: E241
        layout.addWidget(self.logged_in_label,          4,   0,   1,     8)    # noqa: E241
        layout.addWidget(self.logout_button,            4,   8,   1,     4)    # noqa: E241
        layout.addWidget(self.download_spinner,         5,   0,   1,     1)    # noqa: E241
        layout.addWidget(self.download_status_label,    5,   1,   1,    11)    # noqa: E241
        # fmt: on

        self.setLayout(layout)

        self.browse_file_dialog = QFileDialog(None)
        self.browse_file_dialog.setNameFilter("Bin files (*.bin)")

        self.flash_dialog = _FlashDialog(self)
        self.flash_dialog.flash_done.connect(self._flash_done)

        app_model.sig_notify.connect(self._on_app_model_update)

    @property
    def flash_device(self) -> Optional[CommDevice]:
        if self.app_model.connection_interface == ConnectionInterface.SERIAL:
            return self.app_model.serial_connection_device
        if self.app_model.connection_interface == ConnectionInterface.USB:
            return self.app_model.usb_connection_device
        return None

    @property
    def device_name(self) -> Optional[str]:
        if not self.device_selection.isEnabled():
            return None
        else:
            return self.device_name_selection

    def _get_latest_bin_file(self) -> None:
        self.auth_thread = _AuthThread(dev_license=self.dev_license)
        self.auth_thread.started.connect(self._auth_start)
        self.auth_thread.finished.connect(self.auth_thread.deleteLater)
        self.auth_thread.finished.connect(self._auth_stop)
        self.auth_thread.license_loaded.connect(self._license_loaded)
        self.auth_thread.auth_done.connect(self._auth_done)
        self.auth_thread.auth_failed.connect(self._auth_failed)

        self.authenticating = True
        self.auth_thread.start()

    def _auth_start(self) -> None:
        self._show_download_progress("Authenticating...")

    def _auth_stop(self) -> None:
        self._hide_download_progress()
        self.authenticating = False

    def _license_loaded(self, dev_license: DevLicense) -> None:
        self.dev_license = dev_license

    def _auth_done(
        self, auth_info: Tuple[RequestsCookieJar, Tuple[bool, Session, Response]]
    ) -> None:
        self._show_login_status(_LOGGED_IN_MSG)

        cookies, content = auth_info
        self._init_download(cookies, content)

    def _auth_failed(self) -> None:
        login_dialog = _FlashLoginDialog(self)
        authenticated = login_dialog.exec()

        if authenticated:
            self._show_login_status(_LOGGED_IN_MSG)
            cookies, content = login_dialog.get_auth_info()
            if cookies is not None and content is not None:
                self._init_download(cookies, content)

    def _init_download(
        self, cookies: RequestsCookieJar, content: Tuple[bool, Session, Response]
    ) -> None:

        license_agreement_dialog = LicenseAgreementDialog(self.dev_license, self)
        license_accepted = license_agreement_dialog.exec()
        device_download_name = get_flash_download_name(self.flash_device, self.device_name)

        if license_accepted:
            self.download_thread = BinDownloadThread(
                device_download_name, cookies, content[1], content[2]
            )

            self.download_thread.started.connect(self._download_start)
            self.download_thread.finished.connect(self.download_thread.deleteLater)
            self.download_thread.finished.connect(self._download_stop)
            self.download_thread.download_done.connect(self._download_done)
            self.download_thread.download_failed.connect(self._download_failed)

            self.downloading_firmware = True
            self.download_thread.start()

    def _download_start(self) -> None:
        self._show_download_progress("Downloading image file...")

    def _download_stop(self) -> None:
        self._hide_download_progress()
        self.downloading_firmware = False

    def _download_done(self, bin_file: str, version: str) -> None:
        self.bin_file = bin_file
        self._set_version(version)
        self.downloading_firmware = False
        self._draw()

    def _download_failed(self, error_msg: str) -> None:
        self.bin_file = None
        log.error(f"Failed to download firmware: {error_msg}")
        self._draw()

    def _browse_file(self) -> None:
        if self.browse_file_dialog.exec():
            filenames = self.browse_file_dialog.selectedFiles()
            self.bin_file = filenames[0]
            self._reset_download_version()

        self._draw()

    def _show_download_progress(self, text: str) -> None:
        self.download_movie.start()
        self.download_spinner.setHidden(False)
        self.download_status_label.setText(text)
        self.download_status_label.setHidden(False)

    def _hide_download_progress(self) -> None:
        self.download_movie.stop()
        self.download_spinner.setHidden(True)
        self.download_status_label.setHidden(True)
        self.adjustSize()

    def _flash(self) -> None:
        assert self.bin_file is not None

        self.app_model.set_port_updates_pause(True)
        self.flash_dialog.flash(self.bin_file, self.flash_device, self.device_name)

    def _flash_done(self) -> None:
        self.app_model.set_port_updates_pause(False)

    def _on_interface_dd_change(self) -> None:
        self.app_model.set_connection_interface(self.interface_dd.currentData())

    def _on_device_selection_change(self) -> None:
        self.device_name_selection = self.device_selection.currentText()

    def _on_app_model_update(self, app_model: AppModel) -> None:
        if app_model.connection_interface in [ConnectionInterface.SERIAL, ConnectionInterface.USB]:
            interface_index = self.interface_dd.findData(app_model.connection_interface)
            if interface_index == -1:
                raise RuntimeError

            self.interface_dd.setCurrentIndex(interface_index)
            self.stacked.setCurrentIndex(interface_index)

            current_index = self.device_selection.currentIndex()
            if self.flash_device is None or (
                self.flash_device is not None and self.flash_device.name is not None
            ):
                self.device_selection.clear()
                self.device_selection.setEnabled(False)
            else:
                self.device_selection.clear()
                for dev in get_flash_known_devices():
                    self.device_selection.addItem(dev)
                self.device_selection.setEnabled(True)
                if current_index >= 0:
                    self.device_selection.setCurrentIndex(current_index)

            enable_select = not self.authenticating and not self.downloading_firmware
            self.get_latest_button.setEnabled(enable_select)
        else:
            self.device_selection.clear()
            self.device_selection.setEnabled(False)

        self._draw()

    def _draw(self) -> None:
        self.file_label.setText(self.bin_file if self.bin_file else "")
        self.file_label.setEnabled(self.bin_file is not None)
        self.flash_button.setEnabled(
            not self.authenticating
            and not self.downloading_firmware
            and self.flash_device is not None
            and self.bin_file is not None
        )

    def exec(self) -> None:
        if self.app_model.connection_interface == ConnectionInterface.SOCKET:
            self.app_model.set_connection_interface(ConnectionInterface.USB)

        super().exec()

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.authenticating:
            self.auth_thread.terminate()
            self.auth_thread.wait()

        if self.downloading_firmware:
            self.download_thread.terminate()
            self.download_thread.wait()
        super().closeEvent(event)

    def _show_login_status(self, msg: str, disable_logout: bool = False) -> None:
        self.logged_in_label.setText(msg)
        self.logged_in_label.setHidden(False)
        self.logout_button.setHidden(False)
        self.logout_button.setEnabled(not disable_logout)

    def _hide_login_status(self) -> None:
        self.logged_in_label.setHidden(True)
        self.logout_button.setHidden(True)

    def _logout(self) -> None:
        clear_cookies()
        self._show_login_status("Currently not logged in", disable_logout=True)

    def _set_version(self, version: str) -> None:
        self.downloaded_version_label.setText(version)
        self.downloaded_version_label.setEnabled(True)

    def _reset_download_version(self) -> None:
        self.downloaded_version_label.setEnabled(False)
        self.downloaded_version_label.setText("<Downloaded bin file version>")


class FlashButton(QPushButton):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.app_model = app_model

        self.setFixedWidth(100)
        self.setText("Flash")
        self.setIcon(qta.icon("mdi.flash", color=BUTTON_ICON_COLOR))
        self.setToolTip("Flash the device with a bin file")

        app_model.sig_notify.connect(self._on_app_model_update)
        self.pop_up = _FlashPopup(app_model, self)
        self.clicked.connect(self._on_click)

    def _on_click(self) -> None:
        self.pop_up.exec()

    def _on_app_model_update(self, app_model: AppModel) -> None:
        self.setEnabled(app_model.connection_state == ConnectionState.DISCONNECTED)


class _FlashLoginDialog(QDialog):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.cookies: Optional[RequestsCookieJar] = None
        self.content: Optional[Tuple[bool, Session, Response]] = None
        self.logging_in = False
        self.cookies_accepted = False

        self.setWindowTitle("Acconeer developer login")
        self.setMinimumWidth(450)

        self.fetch_prompt = QTextEdit()
        self.fetch_prompt.setPlainText(BIN_FETCH_PROMPT)
        self.fetch_prompt.setReadOnly(True)

        self.usr_edit = QLineEdit()
        self.usr_edit.setPlaceholderText("<Enter email>")
        self.usr_edit.textChanged.connect(self._on_edit_credentials)

        self.pwd_edit = QLineEdit()
        self.pwd_edit.setPlaceholderText("<Enter password>")
        self.pwd_edit.setEchoMode(QLineEdit.EchoMode.Password)
        self.pwd_edit.textChanged.connect(self._on_edit_credentials)

        self.login_status_label = QLabel()
        self.login_status_label.setHidden(True)

        self.remember_me_box = QCheckBox()
        self.remember_me_box.stateChanged.connect(self._on_remember_me_checked)
        self.remember_me_label = QLabel("Remember me")

        self.login_button = QPushButton("Login")
        self.login_button.setDisabled(True)
        self.login_button.clicked.connect(self._handle_login)

        self.cancel_button = QPushButton("Cancel")
        self.cancel_button.clicked.connect(self.reject)

        self.login_spinner = QLabel()
        self.login_spinner.setHidden(True)

        loader_gif = None
        with importlib.resources.path(resources, "loader.gif") as path:
            loader_gif = path

        self.login_movie = QMovie(str(loader_gif))
        self.login_spinner.setMovie(self.login_movie)

        layout = QGridLayout(self)

        # fmt: off
        # Grid layout:                            row, col, rspan, cspan
        layout.addWidget(self.fetch_prompt,       0,   0,   1,     12)    # noqa: E241
        layout.addWidget(self.usr_edit,           1,   0,   1,     12)    # noqa: E241
        layout.addWidget(self.pwd_edit,           2,   0,   1,     12)    # noqa: E241
        layout.addWidget(self.remember_me_box,    3,   0,   1,     1)     # noqa: E241
        layout.addWidget(self.remember_me_label,  3,   1,   1,     11)    # noqa: E241
        layout.addWidget(self.login_status_label, 4,   0,   1,     12)    # noqa: E241
        layout.addWidget(self.login_button,       5,   0,   1,     6)     # noqa: E241
        layout.addWidget(self.login_spinner,      5,   6,   1,     3)     # noqa: E241
        layout.addWidget(self.cancel_button,      5,   9,   1,     3)     # noqa: E241
        # fmt: on

        self.setLayout(layout)

        self.usr_edit.setFocus()

    def _on_edit_credentials(self) -> None:
        if self.usr_edit.text() == "" or self.pwd_edit.text() == "":
            self.login_button.setEnabled(False)
        else:
            self.login_button.setEnabled(True)

    def _on_remember_me_checked(self) -> None:
        if self.remember_me_box.isChecked() and not self.cookies_accepted:
            self.cookies_accepted = CookieConsentDialog(self).exec()
            if not self.cookies_accepted:
                self.remember_me_box.setCheckState(Qt.Unchecked)

    def _handle_login(self) -> None:
        self.login_status_label.setHidden(True)

        usr = self.usr_edit.text()
        pwd = self.pwd_edit.text()

        self.login_thread = _AuthThread(usr, pwd, do_login=True)
        self.login_thread.started.connect(self._login_start)
        self.login_thread.finished.connect(self.login_thread.deleteLater)
        self.login_thread.finished.connect(self._login_stop)
        self.login_thread.auth_done.connect(self._login_done)
        self.login_thread.auth_failed.connect(self._login_failed)

        self.login_thread.start()
        self.logging_in = True

    def _login_start(self) -> None:
        self.login_button.setEnabled(False)
        self._show_login_progress()

    def _login_stop(self) -> None:
        self.logging_in = False
        self._hide_login_progress()
        self.login_button.setEnabled(True)

    def _login_done(
        self, auth_info: Tuple[RequestsCookieJar, Tuple[bool, Session, Response]]
    ) -> None:

        self.cookies, self.content = auth_info

        self.login_status_label.setHidden(True)

        if self.remember_me_box.isChecked():
            save_cookies(self.cookies)

        self.accept()

    def _login_failed(self) -> None:
        self.login_status_label.setText(_WRONG_CREDENTIALS_MSG)
        self.login_status_label.setHidden(False)

    def _show_login_progress(self) -> None:
        self.login_movie.start()
        self.login_spinner.setHidden(False)

    def _hide_login_progress(self) -> None:
        self.login_movie.stop()
        self.login_spinner.setHidden(True)

    def get_auth_info(
        self,
    ) -> (Tuple[Optional[RequestsCookieJar], Optional[Tuple[bool, Session, Response]]]):
        return self.cookies, self.content

    def closeEvent(self, event: QCloseEvent) -> None:
        if self.logging_in:
            self.login_thread.terminate()
            self.login_thread.wait()
        super().closeEvent(event)


class _AuthThread(QThread):
    license_loaded = Signal(DevLicense)
    auth_failed = Signal()
    auth_done = Signal((RequestsCookieJar, Tuple[bool, Session, Response]))

    def __init__(
        self,
        usr: Optional[str] = "",
        pwd: Optional[str] = "",
        do_login: bool = False,
        dev_license: DevLicense = None,
    ) -> None:

        super().__init__()
        self.usr = usr
        self.pwd = pwd
        self.do_login = do_login
        self.dev_license = dev_license

    def run(self) -> None:
        login_succeeded = False

        if self.do_login:
            cookies = login(self.usr, self.pwd)
        else:
            cookies = get_cookies()

        try:
            if cookies is not None:
                content = get_content(cookies)
                login_succeeded = content[0]
            else:
                login_succeeded = False
        except Exception as e:
            log.error(str(e))

        if self.dev_license is not None:
            self.dev_license.load()
            self.license_loaded.emit(self.dev_license)

        if login_succeeded:
            self.auth_done.emit((cookies, content))
        else:
            if not self.do_login:
                clear_cookies()
            self.auth_failed.emit()


class BinDownloadThread(QThread):
    download_done = Signal(str, str)
    download_failed = Signal(str)

    def __init__(
        self,
        device: Optional[str],
        cookies: RequestsCookieJar,
        session: Session,
        response: Response,
    ) -> None:

        super().__init__()
        self.device = device
        self.cookies = cookies
        self.session = session
        self.response = response

    def run(self) -> None:

        if (
            self.device is None
            or self.cookies is None
            or self.session is None
            or self.response is None
        ):
            self.download_failed.emit("Failed to start download")
        else:
            with TemporaryDirectory() as temp_dir:
                try:
                    temp_file, version = download(
                        session=self.session,
                        cookies=self.cookies,
                        path=temp_dir,
                        device=self.device,
                    )

                    if temp_file is not None:
                        bin_file = str(ET_DIR / Path(temp_file).name)
                        os.replace(temp_file, bin_file)
                        self.download_done.emit(bin_file, version)
                    else:
                        self.download_failed.emit(
                            f"Failed to download firmware for device {self.device}"
                        )
                except Exception as e:
                    self.download_failed.emit(str(e))
                finally:
                    self.session.close()


class LicenseAgreementDialog(QDialog):
    def __init__(self, license: DevLicense, parent: QWidget) -> None:
        super().__init__(parent)

        self.setMinimumWidth(800)
        self.setMinimumHeight(600)

        layout = QGridLayout(self)

        self.license_text = QTextEdit(self)
        self.license_text.setReadOnly(True)

        self.accept_box = QCheckBox(self)
        self.accept_box.stateChanged.connect(self._on_accept_check)
        self.accept_label = QLabel("I understand the terms and conditions of the license", self)

        self.accept_button = QPushButton("Accept", self)
        self.accept_button.setEnabled(False)
        self.accept_button.clicked.connect(self.accept)

        self.cancel_button = QPushButton("Cancel", self)
        self.cancel_button.clicked.connect(self.reject)

        # fmt: off
        # Grid layout:                       row, col, rspan, cspan
        layout.addWidget(self.license_text,  0,   0,   10,    12)    # noqa: E241
        layout.addWidget(self.accept_box,    10,  0,   1,     1)     # noqa: E241
        layout.addWidget(self.accept_label,  10,  1,   1,     11)    # noqa: E241
        layout.addWidget(self.accept_button, 11,  0,   1,     6)     # noqa: E241
        layout.addWidget(self.cancel_button, 11,  6,   1,     6)     # noqa: E241
        # fmt: on

        self.setLayout(layout)

        self._set_license_text(license)

    def _on_accept_check(self) -> None:
        self.accept_button.setEnabled(self.accept_box.isChecked())

    def _set_license_text(self, license: DevLicense) -> None:
        self.setWindowTitle(license.get_header())

        self.license_text.insertHtml(license.get_subheader_element())

        content = license.get_content_elements()
        for paragraph in content:
            self.license_text.insertHtml(paragraph)
            self.license_text.insertHtml("<br><br>")

        self.license_text.moveCursor(QTextCursor.MoveOperation.Start)


class CookieConsentDialog(QDialog):
    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent)

        self.setWindowTitle("Cookie consent")

        layout = QGridLayout(self)

        self.consent_label = QLabel(
            "We use cookies to optimize the service of our applications", self
        )
        self.empty_row = QLabel("", self)

        self.cookie_policy_label = QLabel(
            "<a href='https://acconeer.com/cookie-policy-eu/'>Cookie Policy</a>", self
        )
        self.cookie_policy_label.setOpenExternalLinks(True)

        self.privacy_statement_label = QLabel(
            "<a href='https://developer.acconeer.com/privacy-policy/'>Privacy Statement</a>",
            self,
        )
        self.privacy_statement_label.setOpenExternalLinks(True)

        self.accept_button = QPushButton("Accept", self)
        self.accept_button.clicked.connect(self.accept)

        self.deny_button = QPushButton("Deny", self)
        self.deny_button.clicked.connect(self.reject)

        # fmt: off
        # Grid layout:                                 row, col, rspan, cspan
        layout.addWidget(self.consent_label,           0,   0,   1,     4)    # noqa: E241
        layout.addWidget(self.empty_row,               1,   0,   1,     4)    # noqa: E241
        layout.addWidget(self.cookie_policy_label,     2,   0,   1,     4)    # noqa: E241
        layout.addWidget(self.privacy_statement_label, 3,   0,   1,     4)    # noqa: E241
        layout.addWidget(self.accept_button,           4,   0,   1,     2)    # noqa: E241
        layout.addWidget(self.deny_button,             4,   2,   1,     2)    # noqa: E241
        # fmt: on

        self.setLayout(layout)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

from PySide6.QtWidgets import QFrame, QHBoxLayout, QMainWindow, QSplitter, QVBoxLayout, QWidget

from acconeer.exptool.app.new.app_model import AppModel

from .connection_widget import ClientConnectionWidget, GenerationSelection
from .flash_widget import FlashButton
from .misc import ExceptionWidget, HintWidget, VerticalSeparator
from .plugin_widget import PluginControlArea, PluginPlotArea, PluginSelectionArea
from .recording_widget import RecordingWidget
from .status_bar import StatusBar


class MainWindow(QMainWindow):
    def __init__(self, app_model: AppModel) -> None:
        super().__init__()

        self.resize(1280, 720)

        self.setCentralWidget(MainWindowCentralWidget(app_model, self))
        self.setStatusBar(StatusBar(app_model, self))
        self.setWindowTitle("Acconeer Exploration Tool (Beta)")
        self.moveEvent = lambda _: self.saveGeometry()

        app_model.sig_error.connect(self.on_app_model_error)

    def on_app_model_error(self, exception: Exception, traceback_str: Optional[str]) -> None:
        ExceptionWidget(self, exc=exception, traceback_str=traceback_str).exec()


class MainWindowCentralWidget(QWidget):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QVBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        top_bar = TopBar(app_model, self)
        self.layout().addWidget(top_bar)
        self.layout().setStretchFactor(top_bar, 0)

        working_area = WorkingArea(app_model, self)
        self.layout().addWidget(working_area)
        self.layout().setStretchFactor(working_area, 1)


class TopBar(QFrame):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setObjectName("TopBar")
        self.setStyleSheet("QFrame#TopBar {background: #ebebeb;}")

        self.setLayout(QHBoxLayout(self))

        self.layout().addWidget(GenerationSelection(app_model, self))
        self.layout().addWidget(VerticalSeparator(self))
        self.layout().addWidget(ClientConnectionWidget(app_model, self))
        self.layout().addWidget(VerticalSeparator(self))
        self.layout().addWidget(FlashButton(app_model, self))
        self.layout().addWidget(HintWidget(app_model, self))
        self.layout().addStretch(1)
        self.layout().addWidget(RecordingWidget(app_model, self))


class WorkingArea(QSplitter):
    def __init__(self, app_model: AppModel, parent: QWidget) -> None:
        super().__init__(parent)

        self.setLayout(QHBoxLayout(self))
        self.layout().setContentsMargins(0, 0, 0, 0)
        self.layout().setSpacing(0)

        left_area = PluginSelectionArea(app_model, self)
        left_area.setMinimumWidth(250)
        left_area.setMaximumWidth(350)
        self.layout().addWidget(left_area)
        self.layout().setStretchFactor(left_area, 0)

        plot_plugin_area = PluginPlotArea(app_model, self)
        self.layout().addWidget(plot_plugin_area)
        self.layout().setStretchFactor(plot_plugin_area, 1)

        right_area = PluginControlArea(app_model, self)
        right_area.setMinimumWidth(400)
        right_area.setMaximumWidth(450)
        self.layout().addWidget(right_area)
        self.layout().setStretchFactor(right_area, 0)



# Copyright (c) Acconeer AB, 2022
# All rights reserved

from __future__ import annotations

from typing import Optional

import attrs

from acconeer.exptool.a121._core.utils import zip_extended_structures

from ._core import IdleState, Metadata, SensorConfig, SessionConfig


@attrs.frozen
class _SessionPerformanceCalc:
    session_config: SessionConfig = attrs.field()
    extended_metadata: Optional[list[dict[int, Metadata]]] = attrs.field()

    @property
    def update_duration(self) -> float:
        if self.extended_metadata is None:
            raise ValueError("Metadata is None")

        extended_structures = zip_extended_structures(
            self.session_config.groups, self.extended_metadata
        )

        update_duration = 0.0

        for extended_structure in extended_structures:
            update_duration += max(
                _SensorPerformanceCalc(
                    sensor_config, metadata, self.session_config.update_rate
                ).frame_active_duration
                for sensor_config, metadata in extended_structure.values()
            )

        return update_duration

    @property
    def update_rate(self) -> float:
        if self.session_config.update_rate is not None:
            return self.session_config.update_rate
        elif (
            not self.session_config.extended
            and self.session_config.sensor_config.frame_rate is not None
        ):
            return self.session_config.sensor_config.frame_rate
        else:
            return 1.0 / self.update_duration


@attrs.frozen
class _SensorPerformanceCalc:
    """Calculates performance metrics from config and metadata

    The figures are preliminary, approximate, and not guaranteed. Based on conservative
    measurements taken in room temperature.
    """

    sensor_config: SensorConfig = attrs.field()
    metadata: Optional[Metadata] = attrs.field()
    update_rate: Optional[float] = attrs.field()

    @property
    def spf(self) -> int:
        return self.sensor_config.sweeps_per_frame

    @property
    def sweep_rate(self) -> float:
        if self.sensor_config.sweep_rate is not None:
            return self.sensor_config.sweep_rate
        else:
            if self.metadata is None:
                raise ValueError("Metadata is None")
            return self.metadata.max_sweep_rate

    @property
    def sweep_period(self) -> float:
        return 1 / self.sweep_rate

    @property
    def sweep_active_duration(self) -> float:
        if self.metadata is None:
            raise ValueError("Metadata is None")
        return 1 / self.metadata.max_sweep_rate

    @property
    def sweep_idle_duration(self) -> float:
        return self.sweep_period - self.sweep_active_duration

    @property
    def frame_active_duration(self) -> float:
        return self.sweep_period * self.spf

    @property
    def frame_rate(self) -> float:
        max_rate = self.sweep_rate / self.spf

        if self.update_rate is not None:
            rate = self.update_rate
        elif self.sensor_config.frame_rate is not None:
            rate = self.sensor_config.frame_rate
        else:
            rate = max_rate

        return min(rate, max_rate)

    @property
    def frame_period(self) -> float:
        return 1 / self.frame_rate

    @property
    def frame_idle_duration(self) -> float:
        return self.frame_period - self.frame_active_duration

    @property
    def active_current(self) -> float:
        return 85e-3

    @property
    def inter_sweep_idle_current(self) -> float:
        return self._get_idle_current(self.sensor_config.inter_sweep_idle_state)

    @property
    def inter_frame_idle_current(self) -> float:
        return self._get_idle_current(self.sensor_config.inter_frame_idle_state)

    @property
    def sweep_active_charge(self) -> float:
        return self.sweep_active_duration * self.active_current

    @property
    def sweep_idle_charge(self) -> float:
        return self.sweep_idle_duration * self.inter_sweep_idle_current

    @property
    def sweep_charge(self) -> float:
        return self.sweep_active_charge + self.sweep_idle_charge

    @property
    def frame_active_charge(self) -> float:
        return self.sweep_charge * self.spf

    @property
    def frame_idle_charge(self) -> float:
        return self.frame_idle_duration * self.inter_frame_idle_current

    @property
    def frame_charge(self) -> float:
        return self.frame_active_charge + self.frame_idle_charge

    @property
    def average_current(self) -> float:
        return self.frame_charge / self.frame_period

    @staticmethod
    def _get_idle_current(idle_state: IdleState) -> float:
        return {
            IdleState.READY: 70.0e-3,
            IdleState.SLEEP: 1.8e-3,
            IdleState.DEEP_SLEEP: 1.2e-3,
        }[idle_state]



# Copyright (c) Acconeer AB, 2022
# All rights reserved

import shutil
from pathlib import Path
from uuid import uuid4

import platformdirs


ET_DIR = Path(platformdirs.user_data_dir(appname="acconeer_exptool", appauthor="Acconeer AB"))
CODENAME = "plugoneer"
CONFIG_DIR = ET_DIR / CODENAME / "config"
TEMP_DIR = ET_DIR / CODENAME / "temp"


def get_config_dir() -> Path:
    CONFIG_DIR.mkdir(parents=True, exist_ok=True)
    return CONFIG_DIR


def remove_config_dir() -> None:
    shutil.rmtree(CONFIG_DIR, ignore_errors=True)


def get_temp_dir() -> Path:
    TEMP_DIR.mkdir(parents=True, exist_ok=True)
    return TEMP_DIR


def remove_temp_dir() -> None:
    shutil.rmtree(TEMP_DIR, ignore_errors=True)


def get_temp_h5_path() -> Path:
    return (get_temp_dir() / str(uuid4())).with_suffix(".h5")
